import logging

from ppa import appliance

from typing import Union, Optional, Dict, Tuple
from collections import namedtuple

import requests
from requests.compat import urljoin  # type: ignore
from requests.utils import prepend_scheme_if_needed

from . import exceptions
from .query_formatters import assert_formatted_values

logger = logging.getLogger(__name__)


def proxies():
    return {"https": appliance.api_proxy(), "http": appliance.api_proxy()}


def verify():
    return appliance.cert()


class PxmAppliance(namedtuple("PxmAppliance", ["address", "timeout"])):
    @property
    def base_url(self):
        return urljoin(prepend_scheme_if_needed(self.address, "https"), f"pxm_api/v1/")


def _catch_network_errors(response: requests.Response):
    # Some network errors don't cause requests.get to raise, instead
    # they return a response with a status code and text body. Check
    # for those here.
    text = response.text.lower()
    if response.status_code == 502:
        # A 502 (bad gateway) can come from nginx on the PxM Appliance if the rest-api
        # service is down, or it can happen if we can't reach the PxM Appliance at all.
        logger.error(f"502 raw response body {response.text}")
        if all(["nginx" in text, "bad gateway" in text]):
            raise exceptions.NetworkError("(502) Bad Gateway. Failed to connect to the PxM Appliance rest-api service.")
        if "no route to host" in text:
            # The raw message is a bit ugly.
            raise exceptions.NetworkError("(502) No route to host.")
        raise exceptions.NetworkError(f"(502) {response.text}")


def _parse_api_response(response: requests.Response, status_codes: Tuple[int]) -> dict:
    # Parse the JSON response from the PxM Appliance API, then check the status code against
    # what the operation expects. If the status code isn't in the expected list, the JSON
    # body should contain an error from the PxM Appliance API.
    #
    # If there is no JSON response, the request most likely never reached the PxM Appliance API and
    # it could be a network-level error we failed to catch in _check_network_errors.
    try:
        response_body = response.json()
    except ValueError:
        logger.error(f"Could not convert the following response body to JSON: {response.text}")
        raise exceptions.NetworkError(f"({response.status_code}) no valid JSON body in response body '{response.text}'")

    if response.status_code not in status_codes:
        raise exceptions.PxmApiError(
            hint=response_body.get("hint"),
            details=response_body.get("details"),
            code=response_body.get("code"),
            message=response_body.get("message"),
            status_code=response.status_code,
        )
    else:
        return response_body


def _process_server_response(response, status_codes):
    _catch_network_errors(response)
    return _parse_api_response(response, status_codes)


def get(
    pxm: PxmAppliance,
    endpoint: str,
    params: Optional[Union[Dict, None]] = None,
    status_codes: Optional[Tuple[int]] = (200,),
) -> dict:
    url = urljoin(pxm.base_url, endpoint)
    logger.debug(f"sending GET request to {url} with params {params}")
    try:
        response = requests.get(url, params=params, proxies=proxies(), verify=verify(), timeout=pxm.timeout)
    except requests.exceptions.ReadTimeout:
        raise exceptions.NetworkError(f"Connection to {pxm.address}:443 timed out after {pxm.timeout} seconds.")
    return _process_server_response(response, status_codes)


def post(
    pxm: PxmAppliance,
    endpoint: str,
    params: dict = None,
    data: dict = None,
    status_codes: Optional[Tuple[int]] = (200,),
) -> dict:
    url = urljoin(pxm.base_url, endpoint)
    logger.debug(f"sending POST request to {url} with data {data} and params {params}")
    try:
        response = requests.post(url, params=params, data=data, proxies=proxies(), verify=verify(), timeout=pxm.timeout)
    except requests.exceptions.ReadTimeout:
        raise exceptions.NetworkError(f"Connection to {pxm.address}:443 timed out after {pxm.timeout} seconds.")
    return _process_server_response(response, status_codes)


def _all(pxm: PxmAppliance, endpoint: str):
    logger.info(f"getting all {endpoint}")
    return get(pxm, endpoint)


def _search(pxm: PxmAppliance, endpoint: str, **search_params: str):
    assert_formatted_values(**search_params)
    logger.info(f"searching {endpoint} with params {search_params}")
    return get(pxm, endpoint, params=search_params)
