import functools
import logging
import datetime

from typing import Optional, List, Union
import dateutil.parser

from datacaster.classes import CastDataClass

from .exceptions import UnhandledRecordData

logger = logging.getLogger(__name__)


def _optional_datetime(value: Union[str, None]) -> Union[datetime.datetime, None]:
    try:
        return dateutil.parser.isoparse(value)
    except TypeError:
        return None


def _lt(self, other, attribute):
    if isinstance(other, self.__class__):
        return getattr(self, attribute) < getattr(other, attribute)
    raise ValueError(f"Cannot compare objects of {type(self)} with {type(other)}.")


class User(CastDataClass):
    __class_config__ = {
        "cast_functions": {"types": {Optional[datetime.datetime]: _optional_datetime}},
    }

    auth_type: str
    authentication_service_id: Optional[int]
    authentication_service_name: Optional[str]
    delete_in_progress: bool
    email_address: Optional[str]
    enabled: bool
    expires_at: Optional[datetime.datetime]
    id: int
    last_activity: Optional[datetime.datetime]
    name: str
    notes: str
    profiles_count: int
    state: str
    user_groups_count: int
    username: str

    def __lt__(self, other):
        # Users should be sorted by name initially, falling back to ID
        # if for some reason they don't have names. The most common use
        # for sorting will be when generating tables for the task UI, and
        # alphabetical sorting by default makes the most sense.
        #
        # The same logic will be used for all objects coming back from the
        # PxM appliance.
        return _lt(self, other, "username")


class Device(CastDataClass):
    __class_config__ = {
        "cast_functions": {"types": {Optional[datetime.datetime]: _optional_datetime}},
    }

    access_role_ids: Optional[List[int]]
    access_role_names: Optional[List[str]]
    account_source: str
    authentication_service_id: Optional[int]
    authentication_service_name: Optional[str]
    authentication_service_scheme: Optional[str]
    control_account_id: Optional[int]
    device_template_id: int
    device_template_name: str
    device_vendor_name: str
    enabled: bool
    hostname: str
    id: int
    ip_address: str
    last_activity: Optional[datetime.datetime]
    maximum_connections: Optional[int]
    name: str
    notes: str
    profiles_count: int
    traffic_light: str
    updated_at: Optional[datetime.datetime]

    def __lt__(self, other):
        return _lt(self, other, "name")


class ActiveDirectory(CastDataClass):
    __class_config__ = {
        "cast_functions": {"types": {Optional[datetime.datetime]: _optional_datetime}},
    }

    container: str
    control_account_id: Optional[int]
    created_at: Optional[datetime.datetime]
    device_count: int
    domain: str
    enabled: bool
    groups: str  # this is a JSON string containing a list of user groups for some reason!
    hostname: str
    id: int
    name: str
    notes: str
    updated_at: Optional[datetime.datetime]
    user_authentication_service: bool

    def __lt__(self, other):
        return _lt(self, other, "name")


class StaticVault(CastDataClass):
    __class_config__ = {
        "cast_functions": {"types": {Optional[datetime.datetime]: _optional_datetime}},
    }

    control_account_id: Optional[int]
    device_count: int
    created_at: Optional[datetime.datetime]
    id: int
    name: str
    updated_at: Optional[datetime.datetime]

    def __lt__(self, other):
        return _lt(self, other, "name")


class Profile(CastDataClass):
    delete_in_progress: bool
    device_count: int
    device_template_ids: Optional[List[int]]
    groups: list  # This is a list of UserGroup objects.
    id: int
    name: str
    schedule_count: int
    state: str
    superadmin: bool
    task_count: int
    tool_count: int
    user_count: int
    user_group_count: int

    def __lt__(self, other):
        return _lt(self, other, "name")


class UserGroup(CastDataClass):
    __class_config__ = {
        "cast_functions": {"types": {Optional[datetime.datetime]: _optional_datetime}},
    }

    authentication_service_id: Optional[int]
    authentication_service_name: Optional[str]
    enabled: bool
    id: int
    is_synced_group: bool
    name: str
    notes: str
    profile_count: int
    profiles: list  # This is a list of Profile objects.
    synced_at: Optional[datetime.datetime]
    user_count: int

    def __lt__(self, other):
        return _lt(self, other, "name")


class Account(CastDataClass):
    __class_config__ = {
        "cast_functions": {"types": {Optional[datetime.datetime]: _optional_datetime}},
    }

    access_role: Optional[str]
    compromised: bool
    created_at: Optional[datetime.datetime]
    credentials_last_changed_at: Optional[str]
    has_password: bool
    has_sshkey: bool
    id: int
    invalid_at: Optional[datetime.datetime]
    locked: bool
    name: str
    parent_id: int
    parent_type: str
    state: str
    updated_at: Optional[datetime.datetime]

    def __lt__(self, other):
        return _lt(self, other, "name")


class SSHKey(CastDataClass):
    passphrase: Optional[str]
    private: Optional[str]

    def __str__(self):
        return "[* FILTERED *]"

    def __repr__(self):
        return "[* FILTERED *]"


class Secret(CastDataClass):
    account_id: int
    account_username: str
    password: Optional[str]
    ssh_key: Optional[SSHKey]

    def __repr__(self):
        return f"Secret(account_id={self.account_id}, account_username={self.account_username})"

    def __lt__(self, other):
        return _lt(self, other, "account_username")


def create(data_class):
    def _wrapper(func):
        @functools.wraps(func)
        def _create_records(connection, *args, **kwargs) -> list:
            results = func(connection, *args, **kwargs)
            if isinstance(results, list):
                return [data_class(**result) for result in results]
            elif isinstance(results, dict):
                return data_class(**results)
            else:
                logger.error(
                    "Failed to generate PxM object records because the data was not in list or dictionary format."
                )
                logger.error(str(results))
                raise UnhandledRecordData(
                    "Failed to generate PxM object records because the data was not in list or dictionary format."
                )

        return _create_records

    return _wrapper
