from typing import List, Union

from typeguard import typechecked

from . import query_formatters, exceptions, _client, _secrets
from ._records import create, Account, Secret
from ._client import PxmAppliance

ENDPOINT = "accounts"


@typechecked
@create(Account)
def get_all(pxm: PxmAppliance) -> List[Account]:
    """ Get all accounts from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.

    :returns: A list of :ref:`record_types:Account` records.

    .. code-block:: python

        all_accounts = accounts.get_all(pxm)
    """
    return _client._all(pxm, ENDPOINT)


@typechecked
@create(Account)
def search(pxm: PxmAppliance, **search_params: str) -> List[Account]:
    """ Search for PxM Appliance accounts with multiple properties and values.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param search_params: Keyword arguments to use in the search. Each keyword
        should be a property, and each value should be the search value.
        All supplied search values must be formatted by one of the functions in
        :ref:`query_formatters:Query Formatters`.

    :returns: A list of :ref:`record_types:Account` records.

    :raises pxm_appliance.exceptions.ParameterError: if any supplied search value has not been
        formatted by a function in :ref:`query_formatters:Query Formatters`.

    Example - search for all accounts whose parent type is "device" and account name
    starts with "admin".

    .. code-block:: python

        matching_accounts = accounts.search(
            pxm,
            parent_type=query_formatters._eq("device"),
            account=query_formatters._ilike("admin*"),
        )
    .. seealso:: :ref:`index:Search Functions`, :ref:`query_formatters:Query Formatters`
    """
    return _client._search(pxm, ENDPOINT, **search_params)


@typechecked
@create(Account)
def get_by_id(pxm: PxmAppliance, account_id: Union[int, str]) -> Account:
    """ Get an account with a specific ID from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param account_id: The ID of the account.

    :returns: A single :ref:`record_types:Account` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_account = accounts.get_by_id(pxm, 123)
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with id {account_id}."):
        return _client._search(pxm, ENDPOINT, id=query_formatters._eq(account_id))[0]


@typechecked
@create(Account)
def get_by_property(pxm: PxmAppliance, *, name: str, value: Union[int, str]) -> List[Account]:
    """ Find all PxM Appliance accounts whose supplied property matches the
    supplied value. See the code examples for more information.

    This function only accepts a single property and value. The supplied value is case sensitive and
    must be an exact match.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param name: The name of the property to inspect.
    :param value: The value to look for.

    :returns: A list of :ref:`record_types:Account` records.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        local_accounts = accounts.get_by_property(pxm, name="parent_type", value="device")
    """
    return _client._search(pxm, ENDPOINT, **{name: query_formatters._eq(value)})


@typechecked
def reveal(pxm: PxmAppliance, *, account_id: Union[int, str]) -> Secret:
    """ Reveal the secrets of a PxM Appliance account. Specific secrets can be requested by
    supplying a list of names as the `secrets` argument. By default this function will return
    a dictionary containing the keys `account_username` and `password`.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param account_id: The ID of the account to reveal.

    :returns: A :ref:`record_types:Secret` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        secrets = accounts.reveal(pxm, account_id=123)
    """
    account_username, revealed_secrets = _secrets._unpack(_client.post(pxm, f"{ENDPOINT}/{account_id}/secrets"))
    if not revealed_secrets:
        account_record = get_by_id(pxm, account_id)
        if account_record.state in {"UNAPPROVED", "APPROVED"}:
            raise exceptions.NoAccountSecrets(
                f"The supplied account has no secrets as the state is '{account_record.state.lower()}'."
            )
        raise exceptions.NoAccountSecrets("No secrets found for the supplied account.")

    return _secrets._create(account_id, account_username, revealed_secrets)
