from typing import List, Union

from typeguard import typechecked

from . import accounts, exceptions, query_formatters, _client
from ._records import create, ActiveDirectory, Account, Secret
from ._client import PxmAppliance

ENDPOINT = "active-directories"


@typechecked
@create(ActiveDirectory)
def get_all(pxm: PxmAppliance) -> List[ActiveDirectory]:
    """ Get all Active Directories from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.

    :returns: A list of :ref:`record_types:Active Directory` records.

    .. code-block:: python

        all_active_directories = active_directories.get_all(pxm)
    """
    return _client._all(pxm, ENDPOINT)


@typechecked
@create(ActiveDirectory)
def search(pxm: PxmAppliance, **search_params: str) -> List[ActiveDirectory]:
    """ Search for PxM Appliance Active Directories with multiple properties and values.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param search_params: Keyword arguments to use in the search. Each keyword
        should be a property, and each value should be the search value.
        All supplied search values must be formatted by one of the functions in
        :ref:`query_formatters:Query Formatters`.

    :returns: A list of :ref:`record_types:Active Directory` records.

    :raises pxm_appliance.exceptions.ParameterError: if any supplied search value has not been
        formatted by a function in :ref:`query_formatters:Query Formatters`.

    Example - search for all Active Directories whose domain contains "contoso".

    .. code-block:: python

        matching_active_directories = active_directories.search(pxm, domain=query_formatters._ilike("*contoso*"))
    .. seealso:: :ref:`index:Search Functions`, :ref:`query_formatters:Query Formatters`
    """
    return _client._search(pxm, ENDPOINT, **search_params)  # type: ignore


@typechecked
@create(ActiveDirectory)
def get_by_id(pxm: PxmAppliance, active_directory_id: Union[str, int]) -> ActiveDirectory:
    """ Get an Active Directory with a specific ID from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param active_directory_id: The ID of the Active Directory.

    :returns: A single :ref:`record_types:Active Directory` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_active_directories = active_directories.get_by_id(pxm, 123)
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with id {active_directory_id}."):
        return _client._search(pxm, ENDPOINT, id=query_formatters._eq(active_directory_id))[0]


@typechecked
@create(ActiveDirectory)
def get_by_domain(pxm: PxmAppliance, domain: str) -> ActiveDirectory:
    """ Get an Active Directory with a specific domain from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param domain: The Active Directory domain.

    :returns: A single :ref:`record_types:Active Directory` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_active_directories = active_directories.get_by_domain(pxm, "corporate.domain.net")
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with domain {domain}."):
        return _client._search(pxm, ENDPOINT, domain=query_formatters._eq(f'"{domain}"'))[0]


@typechecked
@create(ActiveDirectory)
def get_by_name(pxm: _client.PxmAppliance, name: str) -> ActiveDirectory:
    """ Get a Active Directory with a specific name from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param name: The Active Directory name.

    :returns: A single :ref:`record_types:Active Directory` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_active_directory = active_directories.get_by_name(pxm, "Corporate AD")
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with name {name}."):
        return _client._search(pxm, ENDPOINT, name=query_formatters._eq(name))[0]


@typechecked
def get_control_account_id(pxm: PxmAppliance, *, active_directory_id: Union[int, str]) -> int:
    """ Get the control account ID of the supplied PxM Appliance Active Directory.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param active_directory_id: The ID of the Active Directory.

    :returns: The control account ID of the Active Directory as an int.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        control_account_id = active_directories.get_control_account_id(pxm, 123)
    """
    return get_by_id(pxm, active_directory_id).control_account_id


@typechecked
def get_control_account(pxm: PxmAppliance, *, active_directory_id: Union[str, int]) -> Account:
    """ Get the control account of the supplied PxM Appliance Active Directory.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param active_directory_id: The ID of the Active Directory.

    :returns: A single :ref:`record_types:Account` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        control_account = active_directories.get_control_account(pxm, 123)
    """
    return accounts.get_by_id(pxm, get_control_account_id(pxm, active_directory_id=active_directory_id))  # type: ignore


@typechecked
def reveal_control_account(pxm: PxmAppliance, *, active_directory_id: Union[str, int]) -> Secret:
    """ Reveal the secrets of a PxM Appliance Active Directory control account. Specific
    secrets can be requested by supplying a list of names as the `secrets` argument. By default
    this function will return a dictionary containing the keys `account_username` and
    `password`.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param active_directory_id: The ID of the Active Directory.

    :returns: A :ref:`record_types:Secret` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        secrets = active_directories.reveal_control_account(pxm, 123)
    """
    return accounts.reveal(pxm, account_id=get_control_account_id(pxm, active_directory_id=active_directory_id))


@typechecked
def reveal_account_by_name(
    pxm: _client.PxmAppliance, *, active_directory_id: Union[str, int], account_name: str
) -> Secret:
    """ Reveal the secrets of an Active Directory account by name.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param active_directory_id: The ID of the Active Directory to locate the account in.
    :param account_name: The name of the account to reveal from the Active Directory.

    :returns: A :ref:`record_types:Secret` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        account_secrets = active_directories.reveal_account_by_name(
            pxm,
            active_directory_id=123,
            account_name="my-account"
        )
    """
    with exceptions.raise_on_empty(
        f"Account {account_name} does not exist on Active Directory with ID {active_directory_id}"
    ):
        account = accounts.search(
            pxm,
            parent_type=query_formatters._eq("active-directories"),
            parent_id=query_formatters._eq(active_directory_id),
            name=query_formatters._eq(account_name),
        )[0]
    return accounts.reveal(pxm, account_id=account.id)
