from typing import List, Union

from typeguard import typechecked

from . import _client, _relationships, query_formatters, exceptions
from ._records import create, UserGroup, User, Profile
from ._client import PxmAppliance

ENDPOINT = "user-groups"


@typechecked
@create(UserGroup)
def get_all(pxm: PxmAppliance) -> List[UserGroup]:
    return _client._all(pxm, ENDPOINT)


@typechecked
@create(UserGroup)
def search(pxm: PxmAppliance, **search_params: str) -> List[UserGroup]:
    return _client._search(pxm, ENDPOINT, **search_params)  # type: ignore


@typechecked
@create(UserGroup)
def get_by_id(pxm: PxmAppliance, user_group_id: Union[int, str]) -> UserGroup:
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with id {user_group_id}."):
        return _client._search(pxm, ENDPOINT, id=query_formatters._eq(user_group_id))[0]


@typechecked
def get_profiles(pxm: PxmAppliance, *, user_group_id: Union[str, int]) -> List[Profile]:
    # [
    #     {
    #         "profile": {
    #             "id": 14,
    #             "name": "PPA Compute Testing",
    #             "delete_in_progress": false,
    #             "user_count": 1,
    #             "task_count": 0,
    #             "schedule_count": 0,
    #             "tool_count": 0,
    #             "user_group_count": 6,
    #             "device_count": 2,
    #             "device_template_ids": [
    #                 106
    #             ],
    #             "state": "active",
    #             "superadmin": false
    #         }
    #     }
    # ]
    return [
        Profile(**profile)
        for profile in [
            result["profile"]
            for result in _relationships._user_groups_profiles(
                pxm, user_group_id=query_formatters._eq(user_group_id), select="profile"
            )
        ]
    ]


@typechecked
def get_users(pxm: PxmAppliance, *, user_group_id: Union[str, int]) -> List[User]:
    # [
    #     {
    #         "user": {
    #             "id": 14,
    #             "name": "Tom Guyatt",
    #             "username": "tom.guyatt",
    #             "expires_at": null,
    #             "notes": "",
    #             "enabled": true,
    #             "email_address": "tom.guyatt@osirium.com",
    #             "auth_type": "ad_ldaps",
    #             "authentication_service_id": 1,
    #             "authentication_service_name": "Corporate Domain",
    #             "delete_in_progress": false,
    #             "profiles_count": 6,
    #             "user_groups_count": 4,
    #             "state": "active",
    #             "last_activity": "2020-11-25T11:45:00+00:00"
    #         }
    #     }
    # ]
    return [
        User(**profile)
        for profile in [
            result["user"]
            for result in _relationships._user_groups_users(
                pxm, user_group_id=query_formatters._eq(user_group_id), select="user"
            )
        ]
    ]
