import uuid
import logging

from ppa import ui

logger = logging.getLogger(__name__)


class UnsupportedFileType(Exception):
    pass


def _unpack_binary_file(ppa_file, file_extension):
    temp_path = f"/tmp/{uuid.uuid4()}.{file_extension}"
    logger.info(f"generating temporary {file_extension} file at '{temp_path}'")
    with open(temp_path, "wb") as temp_file:
        temp_file.write(ppa_file.data.encode("raw_unicode_escape"))
    return temp_path


def upload_archive(text: str = None) -> str:
    """ Prompt the user to upload an archive and return the local path
    of the archive file. This function unpacks the data and creates a
    new archive in the /tmp directory of task container. The new filename
    of the stored archive is a UUID.

    Supported archive formats:

        - zip
        - tar
        - gz
        - tgz


    :param text: The text to display to the user when prompting for the file. Defaults
        to `Please upload an archive (supported formats: zip, tar, gz, tgz)`.

    :returns: A 2-tuple containing the original filename, and the path to the temporary
        archive stored on the filesystem.

    :raises UnsupportedFileType: if the archive file uses an unsupported extension.

    Example:

    .. code-block:: python

        upload_filename, archive_path = files.upload_archive("Please upload a tar file")
        ui.output_info(f"Uploaded file {upload_filename} saved to {archive_path}")

        # Do what is required with the file

        os.remove(archive_path)

    .. important:: This function does not clean up the temporary file. Removal after use should be
        done manually as demonstrated above.

    """
    mime_types = {
        "application/x-tar": "tar",
        "application/zip": "zip",
        "application/gzip": "gz",
        "application/x-compressed-tar": "tgz",
    }

    file_types = list(mime_types.values())
    file_types.sort()
    file_types_str = ", ".join(file_types)
    text = text or f"Please upload an archive (supported formats: {file_types_str})"
    uploaded_file = ui.input_file(text, accept=[f".{file_type}" for file_type in file_types])
    logger.info(f"uploaded file: {repr(uploaded_file)}")
    try:
        file_extension = mime_types[uploaded_file.type]
    except KeyError:
        # This shouldn't happen if the accept argument
        # supplied to ui.input_file is correct, but
        # it's better to check anyway.
        raise UnsupportedFileType(f"Uploaded file type {uploaded_file.type} is unsupported.")
    return (uploaded_file.filename, _unpack_binary_file(uploaded_file, file_extension))  # type: ignore
