import logging

from typing import List, Union

from ._client import PxmAppliance
from . import _client, query_formatters

from ._records import Device, UserGroup, Profile

logger = logging.getLogger(__name__)


def _view(pxm: PxmAppliance, view: str, **search_params):
    return _client.get(pxm, view, params=search_params)


def _users_profiles(pxm: PxmAppliance, **search_params):
    return _view(pxm, "users/profiles", **search_params)


def _user_groups_profiles(pxm: PxmAppliance, **search_params):
    return _view(pxm, "user-groups/profiles", **search_params)


def _user_groups_users(pxm: PxmAppliance, **search_params):
    return _view(pxm, "user-groups/users", **search_params)


def _users_user_groups(pxm: PxmAppliance, **search_params):
    return _view(pxm, "users/user-groups", **search_params)


def _devices_user_access(pxm: PxmAppliance, **search_params):
    return _view(pxm, "devices/user-access", **search_params)


def _devices_for_user(pxm: PxmAppliance, user_id: Union[int, str], **search_params: str) -> List[Device]:
    """ Example response from API:
    [
        {
        "device": {
          "id": 13,
          "notes": "",
          "created_at": "2019-01-28T11:35:27.558559",
          "updated_at": "2019-01-28T11:35:31.295619",
          "name": "Windows Server",
          "hostname": "windows-server",
          "enabled": true,
          "device_template_id": 151,
          "maximum_connections": null,
          "control_account_id": 28870,
          "device_template_name": "Windows Server (15.0)",
          "device_vendor_name": "Microsoft",
          "ip_address": "172.16.0.3",
          "access_role_ids": [
            78,
            79
          ],
          "access_role_names": [
            "Administrator",
            "Power User",
            "Guest"
          ],
          "authentication_service_id": null,
          "authentication_service_name": null,
          "authentication_service_scheme": null,
          "account_source": "Local accounts",
          "traffic_light": "usable",
          "profiles_count": 1,
          "last_activity": "2019-01-29T09:56:00+00:00"
        }
      }
    ]
    """
    params = {"user_id": query_formatters._eq(user_id), "select": "device"}
    for key, value in search_params.items():
        params[f"device->>{key}"] = value
    logger.info(f"getting devices for user with id {user_id} using params {params}")
    return [Device(**device_container["device"]) for device_container in _devices_user_access(pxm, **params)]


def _profiles_for_user(pxm: PxmAppliance, user_id: Union[int, str]) -> List[Profile]:
    """ Example response from API:
    [
      {
        "profile": {
          "id": 8,
          "name": "Windows Testing",
          "delete_in_progress": false,
          "user_count": 1,
          "task_count": 0,
          "schedule_count": 0,
          "tool_count": 1,
          "user_group_count": 0,
          "device_count": 1,
          "device_template_ids": [
            151
          ],
          "state": "active",
          "superadmin": false
        }
      }
    ]
    """
    logger.info(f"getting direct profiles for user with id {user_id}")
    return [
        Profile(**profile_container["profile"])
        for profile_container in _users_profiles(pxm, user_id=query_formatters._eq(user_id), select="profile")
    ]


def _profiles_for_user_group(pxm: PxmAppliance, user_group_id: Union[str, int]) -> List[Profile]:
    """ Example response from API:
    [
      {
        "profile": {
          "id": 7,
          "name": "Windows Admins",
          "delete_in_progress": false,
          "user_count": 0,
          "task_count": 1,
          "schedule_count": 0,
          "tool_count": 0,
          "user_group_count": 1,
          "device_count": 10,
          "device_template_ids": [
            2
          ],
          "state": "active",
          "superadmin": false
        }
      }
    ]
    """
    logger.info(f"getting profiles for user group with id {user_group_id}")
    return [
        Profile(**profile_container["profile"])
        for profile_container in _user_groups_profiles(
            pxm, **{"user_group_id": query_formatters._eq(user_group_id), "select": "profile"}
        )
    ]


def _user_groups_for_user(pxm: PxmAppliance, user_id: Union[int, str]) -> List[UserGroup]:
    """ Example response from API:
    [
      {
        "group": {
          "id": 2,
          "name": "Windows Admins",
          "notes": null,
          "enabled": true,
          "is_synced_group": true,
          "synced_at": "2019-05-21T13:56:14.984182",
          "authentication_service_id": 1,
          "authentication_service_name": "dummy.domain.com",
          "profile_count": 0,
          "user_count": 20
        }
      }
    ]
    """
    logger.info(f"getting user groups for user with id {user_id}")
    return [
        UserGroup(**group_container["group"])
        for group_container in _users_user_groups(pxm, **{"user_id": query_formatters._eq(user_id), "select": "group"})
    ]
