from typing import List, Union

from typeguard import typechecked

from . import accounts, exceptions, query_formatters, _client
from ._records import create, Device, Account, Secret

ENDPOINT = "devices"


@typechecked
@create(Device)
def get_all(pxm: _client.PxmAppliance) -> List[Device]:
    """ Get all devices from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.

    :returns: A list of :ref:`record_types:Device` records.

    .. code-block:: python

        all_devices = devices.get_all(pxm)
    """
    return _client._all(pxm, ENDPOINT)


@typechecked
@create(Device)
def search(pxm: _client.PxmAppliance, **search_params: str) -> List[Device]:
    """ Search for PxM Appliance devices with multiple properties and values.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param search_params: Keyword arguments to use in the search. Each keyword
        should be a property, and each value should be the search value.
        All supplied search values must be formatted by one of the functions in
        :ref:`query_formatters:Query Formatters`.

    :returns: A list of :ref:`record_types:Device` records.

    :raises pxm_appliance.exceptions.ParameterError: if any supplied search value has not been
        formatted by a function in :ref:`query_formatters:Query Formatters`.

    Example - search for all enabled devices whose name contains "windows", where the
        template name starts with "Windows Server".

    .. code-block:: python

        matching_users = users.search(
            pxm,
            enabled="true",
            name=query_formatters._ilike("*windows*"),
            device_template_name=query_formatters._ilike("Windows Server*")
        )

    .. seealso:: :ref:`index:Search Functions`, :ref:`query_formatters:Query Formatters`
    """
    return _client._search(pxm, ENDPOINT, **search_params)  # type: ignore


@typechecked
@create(Device)
def get_by_id(pxm: _client.PxmAppliance, device_id: Union[int, str]) -> Device:
    """ Get a device with a specific ID from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param device_id: The ID of the device.

    :returns: A single :ref:`record_types:Device` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_device = devices.get_by_id(pxm, 123)
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with id {device_id}."):
        return _client._search(pxm, ENDPOINT, id=query_formatters._eq(device_id))[0]


@typechecked
@create(Device)
def get_by_hostname(pxm: _client.PxmAppliance, hostname: str) -> Device:
    """ Get a device with a specific hostname from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param hostname: The hostname of the device.

    :returns: A single :ref:`record_types:Device` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_device = devices.get_by_hostname(pxm, "my-server")
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with hostname {hostname}."):
        return _client._search(pxm, ENDPOINT, hostname=query_formatters._eq(hostname))[0]


@typechecked
@create(Device)
def get_by_name(pxm: _client.PxmAppliance, name: str) -> Device:
    """ Get a Device with a specific name from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param name: The Device name.

    :returns: A single :ref:`record_types:Device` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_device = devices.get_by_name(pxm, "Windows 2012")
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with name {name}."):
        return _client._search(pxm, ENDPOINT, name=query_formatters._eq(name))[0]


@typechecked
@create(Device)
def get_by_property(pxm: _client.PxmAppliance, *, name: str, value: Union[int, str]) -> List[Device]:
    """ Find all PxM Appliance users whose supplied property matches the
    supplied value. See the code examples for more information.

    This function only accepts a single property and value. The supplied value is case sensitive
    and must be an exact match.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param name: The name of the property to inspect.
    :param value: The value to look for.

    :returns: A list of :ref:`record_types:Device` records.

    .. code-block:: python

        local_devices = devices.get_by_property(pxm, name="authentication_service_scheme", value="null")
    """
    return _client._search(pxm, ENDPOINT, **{name: query_formatters._eq(value)})


@typechecked
def get_control_account_id(pxm: _client.PxmAppliance, *, device_id: Union[int, str]) -> int:
    """ Get the control account ID of the supplied PxM Appliance device.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param device_id: The ID of the device.

    :returns: The control account ID of the device as an int.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        control_account_id = devices.get_control_account_id(pxm, device_id=123)
    """
    return int(get_by_id(pxm, device_id).control_account_id)


@typechecked
def get_control_account(pxm: _client.PxmAppliance, *, device_id: Union[str, int]) -> Account:
    """ Get the control account of the supplied PxM Appliance device.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param device_id: The ID of the device.

    :returns: A single :ref:`record_types:Account` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        control_account = devices.get_control_account(pxm, device_id=123)
    """
    return accounts.get_by_id(pxm, account_id=get_control_account_id(pxm, device_id=device_id))


@typechecked
def reveal_control_account(pxm: _client.PxmAppliance, *, device_id: Union[str, int]) -> Secret:
    """ Reveal the secrets of a PxM Appliance device control account. Specific secrets
    can be requested by supplying a list of names as the `secrets` argument. By default
    this function will return a dictionary containing the keys `account_username` and
    `password`.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param device_id: The ID of the device.

    :returns: A :ref:`record_types:Secret` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        secrets = devices.reveal_control_account(pxm, device_id=123)
    """
    return accounts.reveal(pxm, account_id=get_control_account_id(pxm, device_id=device_id))


@typechecked
def reveal_account_by_name(pxm: _client.PxmAppliance, *, device_id: Union[str, int], account_name: str) -> Secret:
    """ Reveal the secrets of a device account by name.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param device_id: The ID of the device to locate the account in.
    :param account_name: The name of the account to reveal from the device.

    :returns: A :ref:`record_types:Secret` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        account_secrets = devices.reveal_account_by_name(
            pxm,
            device_id=123,
            account_name="my-account"
        )
    """
    with exceptions.raise_on_empty(f"Account {account_name} does not exist on device with ID {device_id}"):
        account = accounts.search(
            pxm,
            parent_type=query_formatters._eq("devices"),
            parent_id=query_formatters._eq(device_id),
            name=query_formatters._eq(account_name),
        )[0]
    return accounts.reveal(pxm, account_id=account.id)
