import logging

from ppa import ui

from typing import Optional, List, Union

from typeguard import typechecked

from . import users, _relationships, query_formatters
from ._client import PxmAppliance
from ._records import Device

logger = logging.getLogger(__name__)


@typechecked
def user_in_profile(pxm: PxmAppliance, *, user_id: Union[int, str], profile_name: str) -> bool:
    params = {
        "user_id": query_formatters._eq(user_id),
        "profile->>name": query_formatters._eq(profile_name),
        "select": "profile",
    }
    result = bool(_relationships._users_profiles(pxm, **params))
    logger.info(f"user {user_id} in profile '{profile_name}' result {result}")
    return result


@typechecked
def select_user_devices(
    pxm: PxmAppliance,
    *,
    user_id: Optional[Union[int, str]],
    title: Optional[str] = "Select Devices",
    **search_params: str,
) -> List[Device]:
    user = users.get_by_id(pxm, user_id)
    ui.output_progress(
        f"Getting devices for user {user.name} from the PxM Platform...", percent=None, element_id="select_user_devices"
    )
    try:
        devices = users.search_devices(pxm, user_id=user_id, **search_params)
    except Exception as e:
        ui.output_progress(f"Failed to get any devices", percent=0, element_id="select_user_devices")
        raise e

    if not devices:
        ui.output_progress(f"Failed to get any devices", percent=0, element_id="select_user_devices")
        raise Exception(
            f"User '{user.name}' has no profile access to any devices " "matching the search in the PxM Platform."
        )

    ui.output_progress(
        f"Got {len(devices)} devices for user {user.name}", percent=100, element_id="select_user_devices"
    )

    table_data = [
        [
            device.name,  # type: ignore
            device.hostname,  # type: ignore
            device.ip_address,  # type: ignore
            str(device.enabled),  # type: ignore
            device.device_template_name,  # type: ignore
            device.traffic_light.title(),  # type: ignore
        ]
        for device in devices
    ]

    indices: List[int] = ui.input_table(
        title, header=["Device Name", "Hostname", "Address", "Enabled", "Template Name", "Status"], data=table_data
    )
    indices.sort()

    return [devices[i] for i in indices]
