from typeguard import typechecked

from typing import List, Union

from ._records import create, StaticVault, Secret
from ._client import PxmAppliance
from . import _client, accounts, exceptions, query_formatters

ENDPOINT = "static-vaults"


@typechecked
@create(StaticVault)
def get_all(pxm: PxmAppliance) -> List[StaticVault]:
    """ Get all Static Vaults from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.

    :returns: A list of :ref:`record_types:Static Vault` records.

    .. code-block:: python

        all_static_vaults = static_vaults.get_all(pxm)
    """
    return _client._all(pxm, ENDPOINT)


@typechecked
@create(StaticVault)
def search(pxm: PxmAppliance, **search_params: str) -> List[StaticVault]:
    """ Search for PxM Appliance Static Vaults with multiple properties and values.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param search_params: Keyword arguments to use in the search. Each keyword
        should be a property, and each value should be the search value.
        All supplied search values must be formatted by one of the functions in
        :ref:`query_formatters:Query Formatters`.

    :returns: A list of :ref:`record_types:Static Vault` records.

    :raises pxm_appliance.exceptions.ParameterError: if any supplied search value has not been
        formatted by a function in :ref:`query_formatters:Query Formatters`.

    Example - search for all Static Vaults whose name contains "static".

    .. code-block:: python

        matching_static_vaults = static_vaults.search(pxm, name=query_formatters._ilike("*static*"))
    .. seealso:: :ref:`index:Search Functions`, :ref:`query_formatters:Query Formatters`
    """
    return _client._search(pxm, ENDPOINT, **search_params)  # type: ignore


@typechecked
@create(StaticVault)
def get_by_id(pxm: PxmAppliance, static_vault_id: Union[str, int]) -> StaticVault:
    """ Get a Static Vault with a specific ID from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param static_vault_id: The ID of the Static Vault.

    :returns: A single :ref:`record_types:Static Vault` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_static_vault = static_vaults.get_by_id(pxm, 123)
    """

    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with id {static_vault_id}."):
        return _client._search(pxm, ENDPOINT, id=query_formatters._eq(static_vault_id))[0]


@typechecked
@create(StaticVault)
def get_by_name(pxm: PxmAppliance, name: str) -> StaticVault:
    """ Get a Static Vault with a specific name from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param name: The Static Vault name.

    :returns: A single :ref:`record_types:Static Vault` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_static_vault = static_vaults.get_by_name(pxm, "static")
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with name {name}."):
        return _client._search(pxm, ENDPOINT, name=query_formatters._eq(name))[0]


@typechecked
def reveal_account_by_name(pxm: PxmAppliance, *, vault_id: Union[str, int], account_name: str) -> Secret:
    """ Reveal the secrets of a Static Vault account by name.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param vault_id: The name of the static vault to locate the account in.
    :param account_name: The name of the account to reveal from the static vault.

    :returns: A :ref:`record_types:Secret` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        account_secrets = static_vaults.reveal_account_by_name(
            pxm,
            vault_id=123,
            account_name="my-account"
        )
    """
    with exceptions.raise_on_empty(f"Account {account_name} does not exist in Static Vault with ID {vault_id}"):
        account = accounts.search(
            pxm,
            parent_type=query_formatters._eq("static-vaults"),
            parent_id=query_formatters._eq(vault_id),
            name=query_formatters._eq(account_name),
        )[0]
    return accounts.reveal(pxm, account_id=account.id)
