from itertools import zip_longest
from typing import List, Union

from typeguard import typechecked

from . import _client, query_formatters, exceptions, _relationships
from ._records import create, User, Device, Profile, UserGroup
from ._client import PxmAppliance

ENDPOINT = "users"


@typechecked
@create(User)
def get_all(pxm: PxmAppliance) -> List[User]:
    """ Get all users from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.

    :returns: A list of :ref:`record_types:User` records.

    .. code-block:: python

        all_users = users.get_all(pxm)
    """
    return _client._all(pxm, ENDPOINT)


@typechecked
@create(User)
def search(pxm: PxmAppliance, **search_params: str) -> List[User]:
    """ Search for PxM Appliance users with multiple properties and values.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param search_params: Keyword arguments to use in the search. Each keyword
        should be a property, and each value should be the search value.
        All supplied search values must be formatted by one of the functions in
        :ref:`query_formatters:Query Formatters`.

    :returns: A list of :ref:`record_types:User` records.

    :raises pxm_appliance.exceptions.ParameterError: if any supplied search value has not been
        formatted by a function in :ref:`query_formatters:Query Formatters`.

    Example - search for all enabled, locally authenticated users, whose username contains "smith".

    .. code-block:: python

        matching_users = users.search(
            pxm,
            auth_type=query_formatters._eq("password"),
            enabled=query_formatters._eq("true"),
            username=query_formatters._ilike("*smith*")
        )
    .. seealso:: :ref:`index:Search Functions`, :ref:`query_formatters:Query Formatters`
    """
    return _client._search(pxm, ENDPOINT, **search_params)  # type: ignore


@typechecked
@create(User)
def get_by_id(pxm: PxmAppliance, user_id: Union[int, str]) -> User:
    """ Get a user with a specific ID from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param user_id: The ID of the user.

    :returns: A single :ref:`record_types:User` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_user = users.get_by_id(pxm, 123)
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with id {user_id}."):
        return _client._search(pxm, ENDPOINT, id=query_formatters._eq(user_id))[0]


@typechecked
@create(User)
def get_by_username(pxm: PxmAppliance, username: str) -> User:
    """ Get a user with a specific username from the PxM Appliance.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param username: The username of the user.

    :returns: A single :ref:`record_types:User` record.

    :raises pxm_appliance.exceptions.NoSuchObject: if no record is found.

    .. code-block:: python

        my_user = users.get_by_username(pxm, "my.user")
    """
    with exceptions.raise_on_empty(f"No {ENDPOINT} record found on the PxM Platform with username {username}."):
        return _client._search(pxm, ENDPOINT, username=query_formatters._eq(username))[0]


@typechecked
@create(User)
def get_by_property(pxm: PxmAppliance, *, name: str, value: str) -> List[User]:
    """ Find all PxM Appliance users whose supplied property matches the
    supplied value. See the code examples for more information.

    This function only accepts a single property and value. The supplied value is case sensitive
    and must be an exact match.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param name: The name of the property to inspect.
    :param value: The value to look for.

    :returns: A list of :ref:`record_types:User` records.

    .. code-block:: python

        local_users = users.get_by_property(pxm, name="auth_type", value="password")
        enabled_users = users.get_by_property(pxm, name="enabled", value="true")
    """
    return _client._search(pxm, ENDPOINT, **{name: query_formatters._eq(value)})


@typechecked
def get_all_devices(pxm: PxmAppliance, *, user_id: Union[str, int]) -> List[Device]:
    """ Get the list of all devices a certain PxM Appliance user has access to.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param user_id: The ID of the user.

    :returns: A list of :ref:`record_types:Device` records.

    .. code-block:: python

        my_devices = users.get_all_devices(
            pxm,
            user_id=users.get_by_username(pxm, "my.user").id
        )
    """
    return _relationships._devices_for_user(pxm, user_id)


@typechecked
def search_devices(pxm: PxmAppliance, *, user_id: Union[str, int], **search_params: str) -> List[Device]:
    """ Search the list of all devices a PxM Appliance user has access
    to and return the search results.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param user_id: The ID of the user.
    :param search_params: Keyword arguments to use in the search. Each keyword
        should be a property, and each value should be the search value.
        All supplied search values must be formatted by one of the functions in
        :ref:`query_formatters:Query Formatters`.

    :returns: A list of :ref:`record_types:Device` records.

    .. code-block:: python

        my_windows_devices = users.search_devices(
            pxm,
            user_id=users.by_username(pxm, "my.user").id,
            vendor="microsoft",
            enabled="true"
        )
    """
    return _relationships._devices_for_user(pxm, user_id, **search_params)


@typechecked
def get_user_groups(pxm: PxmAppliance, *, user_id: Union[str, int]) -> List[UserGroup]:
    """ Get a list of user groups the supplied PxM Appliance user is a member of.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param user_id: The ID of the user.

    :returns: A list of :ref:`record_types:User Group` records.

    .. code-block:: python

        my_user_groups = users.get_user_groups(
            pxm, user_id=users.get_by_username(pxm, "my.user").id
        )
    """
    return _relationships._user_groups_for_user(pxm, user_id)


@typechecked
def get_direct_profiles(pxm: PxmAppliance, *, user_id: Union[str, int]) -> List[Profile]:
    """ Get a list of PxM Appliance profiles the supplied user has direct access to.

    Direct access means the user is added directly into the profile, not via a user group.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param user_id: The ID of the user.

    :returns: A list of :ref:`record_types:Profile` records.

    .. code-block:: python

        my_direct_profiles = users.get_direct_profiles(
            user_id=users.get_by_username(pxm, "my.user").id
        )
    """
    return _relationships._profiles_for_user(pxm, user_id=user_id)


@typechecked
def get_user_group_profiles(pxm: PxmAppliance, *, user_id: Union[str, int]) -> List[Profile]:
    """ Get a list of PxM Appliance profiles the supplied user has access to via a user group.

    User group access means the user is a member of one or more user groups added to the profile. Direct
    profile access is ignored by this function.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param user_id: The ID of the user.

    :returns: A list of :ref:`record_types:Profile` records.

    .. code-block:: python

        my_user_group_profiles = users.get_user_group_profiles(
            user_id=users.get_by_username(pxm, "my.user").id
        )
    """
    profiles = []
    for group in get_user_groups(pxm, user_id=user_id):
        profiles.extend(_relationships._profiles_for_user_group(pxm, group.id))  # type: ignore
    return profiles  # type: ignore


@typechecked
def get_all_profiles(pxm: PxmAppliance, *, user_id: Union[int, str]) -> List[Profile]:
    """ Get a list of all PxM Appliance profiles the supplied user has access to. This
    includes both direct access and user group access.

    :param pxm: An :ref:`record_types:Appliance Adaptor`.
    :param user_id: The ID of the user.

    :returns: A list of Pxm Appliance :ref:`record_types:Profile` records.

    .. code-block:: python

        all_my_profiles = users.get_all_profiles(
            user_id=users.get_by_username(pxm, "my.user").id
        )
    """
    profiles = []
    for direct_profile, group_profile in zip_longest(
        get_direct_profiles(pxm, user_id=user_id), get_user_group_profiles(pxm, user_id=user_id)
    ):
        # If the user has both direct access *and* user group
        # access to a profile, just count that profile once.
        if direct_profile == group_profile:
            profiles.append(direct_profile)
        else:
            # Otherwise count both, provided they are not
            # a None value generated by zip_longest.
            if direct_profile:
                profiles.append(direct_profile)
            if group_profile:
                profiles.append(group_profile)

    return profiles
