import logging
import urllib3
import requests
import functools
import contextlib

from typing import Callable, List

from pxm_appliance import client, query_formatters

from . import exceptions
from .records import Query


logger = logging.getLogger(__name__)


def _is(value: None) -> query_formatters.FormattedValue:
    return query_formatters.FormattedValue(f"is.{value if value is not None else 'null'}", value)


def _not(value: None) -> query_formatters.FormattedValue:
    return query_formatters._not(_is(value))


QUERY_FORMATTERS = {
    "equal": query_formatters._eq,
    "not_equal": query_formatters._neq,
    "like": query_formatters._ilike,
    "greater_than": query_formatters._gt,
    "less_than": query_formatters._lt,
    "is": _is,
    "not": _not,
}


@contextlib.contextmanager
def get_client(pam_address: str):
    yield client.initialize(address=pam_address)


def search(func: Callable) -> Callable:
    @functools.wraps(func)
    def _search_operation(pam_address: str, queries: List[Query]):
        if list(filter(lambda query: sorted(list(query.keys())) != ["attribute", "query", "value"], queries)):
            raise exceptions.InvalidQuery("Query dictionaries must only contain 'attribute', 'query', & 'value' keys.")

        if invalid_query_types := [query["query"] for query in queries if query["query"] not in QUERY_FORMATTERS]:
            raise exceptions.InvalidQuery(
                f"The following query types are invalid: '{invalid_query_types}'. "
                f"Valid queries are '{list(QUERY_FORMATTERS.keys())}'."
            )

        formatted_queries = {}
        for query in queries:
            value = query["value"]
            query_type = query["query"]
            attribute = query["attribute"]
            if all([query_type in {"is", "not"}, not isinstance(value, type(None))]):
                raise exceptions.InvalidQuery(f"The value of 'is' & 'not' queries must be null. Received '{type(value)}'.")
            if all([query_type == "like", not isinstance(value, str)]):
                raise exceptions.InvalidQuery(f"The value of 'like' queries must be a string. Received '{type(value)}'.")
            if all([query_type in {"less_than", "greater_than"}, not isinstance(value, int)]):
                raise exceptions.InvalidQuery(
                    f"The value of 'less_than' or 'greater_than' queries must be an integer. Received '{type(value)}'."
                )
            if query_type in {"equal", "not_equal"}:
                if not isinstance(value, (str, int, bool)):
                    value = f'"{value}"'
                else:
                    pass
            formatted_queries[attribute] = QUERY_FORMATTERS[query_type](value)
        logger.debug(f"{formatted_queries=}")

        return func(pam_address, queries=formatted_queries)

    return _search_operation


def api_call(func: Callable) -> Callable:
    @functools.wraps(func)
    def _inner(pam_address: str, *args, **kwargs):
        try:
            # If one decorated function calls another decorated function, we don't want to open
            # a second client. If the first argument isn't a string, the client was already opened.
            if not isinstance(pam_address, str):
                return func(pam_address, *args, **kwargs)
            with get_client(pam_address) as api_client:
                return func(api_client, *args, **kwargs)
        except (urllib3.exceptions.MaxRetryError, requests.exceptions.ProxyError):
            raise exceptions.PAMVaultNotFound(f"No PAM vault has been created in PPA with IP/DNS address '{pam_address}'.")

    return _inner


def record_list(record_type, records):
    return list(map(lambda record: record_type(**record.__dict__).__dict__, records))


def single_record(record_type, record):
    return record_type(**record.__dict__).__dict__


def get_cell_value(item, key):
    return str(item.get(key)) if item.get(key) is not None else ""
