"""
Accounts

This module contains actions for reading account information from PAM.

!!! question "Supported Versions"

    This plugin supports PAM versions 6.5.0 & newer.

!!! danger "Remember"

    You must have a PAM Appliance provisioned as a _Vault_ inside PPA to use this plugin.
"""
import uuid

from typing import List, Union, Optional

from ppa import ui
from pxm_appliance import accounts
from playbook_tools import tables, process

from ._api import api_call, search, record_list, single_record, get_cell_value
from .records import Account, AccountData, Secret, SecretData, Query


TABLE_HEADER = ["Name", "Parent Type", "State", "Credentials Last Updated At"]
TABLE_FIELDS = ["name", "parent_type", "state", "credentials_last_changed_at"]


@api_call
def get_all(pam_address: str) -> List[Account]:
    """Get all accounts from the PAM appliance.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

??? success "Output"

    A list of [Accounts](../inputs_outputs/#accounts).

??? example
    ```YAML linenums="1"
    - pam.accounts.get_all:
        pam_address: pam.internal.net
      save: all_accounts
    ```
"""
    return record_list(AccountData, accounts.get_all(pam_address))


@api_call
@search
def search(pam_address: str, queries: List[Query]) -> process.SearchResult:
    """Search for accounts in the PAM appliance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __queries:__ a list of [Query](../inputs_outputs/#query) dictionaries to use in the search

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [Accounts](../inputs_outputs/#account) found

    - **all**: a list of [Accounts](../inputs_outputs/#account) found

    - **first**: the first [Account](../inputs_outputs/#account) found

    - **last**: the last [Account](../inputs_outputs/#account) found

??? warning "Search Result Format"

    The search result format was updated in version 2.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

??? example

    Finding all known accounts in PAM:

    ```YAML linenums="1"
    - pam.accounts.search:
        pam_address: pam.internal.net
        queries:
          - attribute: state
            query: equal
            value: Known
      save: search_results
    ```

!!! tip "Wildcard Queries"

    You can use the ___%___ character as a wildcard in the `value` of each query.

"""
    return process.search_result(record_list(AccountData, accounts.search(pam_address, **queries)))


@api_call
def get_by_id(pam_address: str, account_id: Union[str, int]) -> Account:
    """Get an account using its ID.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    _account_id:__ the ID of the account

??? success "Output"

    A single [Account](../inputs_outputs/#account).

??? example

    ```YAML linenums="1"
    - pam.accounts.get_by_id:
        pam_address: pam.internal.net
        account_id: 10
      save: account
    ```
"""
    return single_record(AccountData, accounts.get_by_id(pam_address, account_id))


@api_call
def reveal(pam_address: str, account_id: Union[str, int]) -> Secret:
    """Reveal an account's secrets using its ID.

!!! question "Minimum Plugin Version: 0.1.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __account_id:__ the ID of the account

??? success "Output"

    A single [Secret](../inputs_outputs/#secret).

??? example

    ```YAML linenums="1"
    - pam.accounts.reveal:
        pam_address: pam.internal.net
        account_id: 3
      save: secret
    ```
"""
    return single_record(SecretData, accounts.reveal(pam_address, account_id=account_id))


def output_table(text: str, accounts: Union[Account, List[Account]]):
    """Display accounts in a table.

The table will have the following columns:

- Name
- Parent Type
- State
- Credentials Last Updated At

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __accounts:__ any number of [Accounts](../inputs_outputs/#account)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.accounts.get_all:
        pam_address: pam.internal.net
      save: all_accounts

    - pam.accounts.output_table:
        text: All Accounts
      load:
        accounts: all_accounts
    ```
    """
    accounts = [accounts] if not isinstance(accounts, list) else accounts
    ui.output_table(
        text, header=TABLE_HEADER, data=[[get_cell_value(account, field) for field in TABLE_FIELDS] for account in accounts],
    )


def display(text: str, accounts: Union[Account, List[Account]]) -> None:
    """Display accounts in a table.

The table will have the following columns:

- Name
- Parent Type
- State
- Credentials Last Updated At

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __accounts:__ any number of [Accounts](../inputs_outputs/#account)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.accounts.get_all:
        pam_address: pam.internal.net
      save: all_accounts

    - pam.accounts.display:
        text: All Accounts
      load:
        accounts: all_accounts
    ```
    """
    tables.output_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=accounts)


def input_table(
    text: str, accounts: Union[Account, List[Account]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Account]:
    """Display accounts in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- Parent Type
- State
- Credentials Last Updated At

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __accounts:__ any number of [Accounts](../inputs_outputs/#account)

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [Accounts](../inputs_outputs/#accounts).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.

    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number of selections.

??? example

    ```YAML linenums="1"
    - pam.accounts.get_all:
        pam_address: pam.internal.net
      save: all_accounts

    - pam.accounts.input_table:
        text: Select an Account
        minimum: 1
        maximum: 1
      load:
        accounts: all_accounts
      save: selected_accounts
    ```
    """
    accounts = [accounts] if not isinstance(accounts, list) else accounts
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())

    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[get_cell_value(account, field) for field in TABLE_FIELDS] for account in accounts],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [accounts[i] for i in indices]


def select(
    text: str, accounts: Union[Account, List[Account]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display accounts in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Parent Type
- State
- Credentials Last Updated At

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __accounts:__ any number of [Accounts](../inputs_outputs/#account)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Accounts](../inputs_outputs/#account)

    - **all**: a list of selected [Accounts](../inputs_outputs/#account)

    - **first**: the first selected [Account](../inputs_outputs/#account)

??? example

    ```YAML linenums="1"
    - pam.accounts.get_all:
        pam_address: pam.internal.net
      save: all_accounts

    - pam.accounts.select:
        text: Select Accounts
      load:
        accounts: all_accounts
      save: selection
    ```
    """
    return tables.input_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=accounts, minimum=minimum, maximum=maximum)


def select_one(text: str, accounts: Union[Account, List[Account]]) -> Account:
    """Display accounts in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Parent Type
- State
- Credentials Last Updated At

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __accounts:__ any number of [Accounts](../inputs_outputs/#account)

??? success "Output"

    A single [Account](../inputs_outputs/#account).

??? example

    ```YAML linenums="1"
    - pam.accounts.get_all:
        pam_address: pam.internal.net
      save: all_accounts

    - pam.accounts.select_one:
        text: Select an Account
      load:
        accounts: all_accounts
      save: account
    ```
    """
    return select(text, accounts, minimum=1, maximum=1)["first"]
