"""
Active Directories

This module contains actions for reading Active Directory information from PAM.

!!! question "Supported Versions"

    This plugin supports PAM versions 6.5.0 & newer.

!!! danger "Remember"

    You must have a PAM Appliance provisioned as a _Vault_ inside PPA to use this plugin.
"""

import uuid

from typing import List, Optional, Union

from ppa import ui
from pxm_appliance import active_directories
from playbook_tools import tables, process

from ._api import api_call, search, single_record, record_list, get_cell_value
from .records import ActiveDirectory, ActiveDirectoryData, Secret, SecretData, Query

TABLE_HEADER = ["Name", "Domain", "Host Name", "Device Count", "User Authentication Service"]
TABLE_FIELDS = [
    "name",
    "domain",
    "hostname",
    "device_count",
    "user_authentication_service",
]


@api_call
def get_all(pam_address: str) -> List[ActiveDirectory]:
    """Get all Active Directories from the PAM appliance.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

??? success "Output"

    A list of [ActiveDirectories](../inputs_outputs/#activedirectory).

??? example
    ```YAML linenums="1"
    - pam.active_directories.get_all:
        pam_address: pam.internal.net
      save: all_active_directories
    ```
"""
    return record_list(ActiveDirectoryData, active_directories.get_all(pam_address))


@api_call
@search
def search(pam_address: str, queries: List[Query]) -> process.SearchResult:
    """Search for Active Directories in the PAM appliance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __queries:__ a list of [Queries](../inputs_outputs/#query) to use in the search

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [ActiveDirectories](../inputs_outputs/#activedirectory) found

    - **all**: a list of [ActiveDirectories](../inputs_outputs/#activedirectory) found

    - **first**: the first [ActiveDirectory](../inputs_outputs/#activedirectory) found

    - **last**: the last [ActiveDirectory](../inputs_outputs/#activedirectory) found

??? warning "Search Result Format"

    The search result format was updated in version 2.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

!!! tip "Wildcard Queries"

    You can use the ___%___ character as a wildcard in the `value` of each query.

??? example

    Finding the Active Directory used as the User Authentication Service in PAM:

    ```YAML linenums="1"
    - pam.active_directories.search:
        pam_address: pam.internal.net
        queries:
          - attribute: user_authentication_service
            query: equal
            value: true
      save: search_results
    ```

"""
    return process.search_result(record_list(ActiveDirectoryData, active_directories.search(pam_address, **queries)))


@api_call
def get_by_id(pam_address: str, active_directory_id: Union[str, int]) -> ActiveDirectory:
    """Get an Active Directory using its ID.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __active_directory_id:__ the ID of the Active Directory

??? success "Output"

    A single [ActiveDirectory](../inputs_outputs/#activedirectory).

??? example

    ```YAML linenums="1"
    - pam.active_directories.get_by_id:
        pam_address: pam.internal.net
        active_directory_id: 10
      save: active_directory
    ```
"""
    return single_record(ActiveDirectoryData, active_directories.get_by_id(pam_address, active_directory_id))


@api_call
def get_by_domain(pam_address: str, domain: str) -> ActiveDirectory:
    """Get an Active Directory using its FQDN.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __domain:__ the Active Directory FQDN

??? success "Output"

    A single [ActiveDirectory](../inputs_outputs/#activedirectory).

??? example

    ```YAML linenums="1"
    - pam.active_directories.get_by_domain:
        pam_address: pam.internal.net
        domain: corporate.domain.net
      save: active_directory
    ```
"""
    return single_record(ActiveDirectoryData, active_directories.get_by_domain(pam_address, domain))


@api_call
def get_by_name(pam_address: str, name: str) -> ActiveDirectory:
    """Get an Active Directory using its name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __name:__ the name of the Active Directory

??? success "Output"

    A single [ActiveDirectory](../inputs_outputs/#activedirectory).

??? example

    ```YAML linenums="1"
    - pam.active_directories.get_by_name:
        pam_address: pam.internal.net
        name: Corporate Domain
      save: active_directory
    ```
"""
    return single_record(ActiveDirectoryData, active_directories.get_by_name(pam_address, name))


@api_call
def reveal_control_account(pam_address: str, active_directory_id: Union[str, int]) -> Secret:
    """Reveal an Active Directory's control account secrets using its ID.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __active_directory_id:__ the ID of the Active Directory

??? success "Output"

    A single [Secret](../inputs_outputs/#secret).

??? example

    ```YAML linenums="1"
    - pam.active_directories.reveal_control_account:
        pam_address: pam.internal.net
        active_directory_id: 3
      save: active_directory_secrets
    ```
"""
    return single_record(
        SecretData, active_directories.reveal_control_account(pam_address, active_directory_id=active_directory_id)
    )


@api_call
def reveal_account_by_name(pam_address: str, active_directory_id: Union[str, int], account_name: str) -> Secret:
    """Reveal an Active Directory account's secrets using the account name.

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __active_directory_id:__ the ID of the Active Directory

    __account_name:__ the name of the Active Directory account

??? success "Output"

    A single [Secret](../inputs_outputs/#secret).

??? example

    ```YAML linenums="1"
    - pam.active_directories.reveal_account_by_name:
        pam_address: pam.internal.net
        active_directory_id: 3
        account_name: ppa_service_account
      save: secret
    ```
"""
    return single_record(
        SecretData,
        active_directories.reveal_account_by_name(
            pam_address, active_directory_id=active_directory_id, account_name=account_name
        ),
    )


def output_table(text: str, active_directories: Union[ActiveDirectory, List[ActiveDirectory]]):
    """Display Active Directories in a table.

The table will have the following columns:

- Name
- Domain
- Host Name
- Device Count
- User Authentication Service

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __active_directories:__ any number of [ActiveDirectories](../inputs_outputs/#activedirectory)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.active_directories.get_all:
        pam_address: pam.internal.net
      save: active_directories

    - pam.active_directories.output_table:
        text: All Active Directories
      load:
        active_directories: active_directories
    ```
    """
    active_directories = [active_directories] if not isinstance(active_directories, list) else active_directories
    ui.output_table(
        text,
        header=TABLE_HEADER,
        data=[[get_cell_value(active_directory, field) for field in TABLE_FIELDS] for active_directory in active_directories],
    )


def display(text: str, active_directories: Union[ActiveDirectory, List[ActiveDirectory]]) -> None:
    """Display Active Directories in a table.

The table will have the following columns:

- Name
- Domain
- Host Name
- Device Count
- User Authentication Service

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __active_directories:__ any number of [ActiveDirectories](../inputs_outputs/#activedirectory)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.active_directories.get_all:
        pam_address: pam.internal.net
      save: active_directories

    - pam.active_directories.display:
        text: All Active Directories
      load:
        active_directories: active_directories
    ```
    """
    tables.output_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=active_directories)


def input_table(
    text: str,
    active_directories: Union[ActiveDirectory, List[ActiveDirectory]],
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
) -> List[ActiveDirectory]:
    """Display Active Directories in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- Domain
- Host Name
- Device Count
- User Authentication Service

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __active_directories:__ any number of [ActiveDirectories](../inputs_outputs/#activedirectory)

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [ActiveDirectories](../inputs_outputs/#activedirectory).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.

    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number of selections.

??? example

    ```YAML linenums="1"
    - pam.active_directories.get_all:
        pam_address: pam.internal.net
      save: active_directories

    - pam.active_directories.input_table:
        text: Select an Active Directory
        minimum: 1
        maximum: 1
      load:
        active_directories: active_directories
      save: selected_active_directory
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    active_directories = [active_directories] if not isinstance(active_directories, list) else active_directories

    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[get_cell_value(active_directory, field) for field in TABLE_FIELDS] for active_directory in active_directories],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [active_directories[i] for i in indices]


def select(
    text: str,
    active_directories: Union[ActiveDirectory, List[ActiveDirectory]],
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
) -> tables.TableSelection:
    """Display Active Directories in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Domain
- Host Name
- Device Count
- User Authentication Service

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __active_directories:__ any number of [ActiveDirectories](../inputs_outputs/#activedirectory)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [ActiveDirectories](../inputs_outputs/#activedirectory)

    - **all**: a list of selected [ActiveDirectories](../inputs_outputs/#activedirectory)

    - **first**: the first selected [ActiveDirectory](../inputs_outputs/#activedirectory)

??? example

    ```YAML linenums="1"
    - pam.active_directories.get_all:
        pam_address: pam.internal.net
      save: active_directories

    - pam.active_directories.select:
        text: Select Active Directories
      load:
        active_directories: active_directories
      save: selection
    ```
    """
    return tables.input_table(
        text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=active_directories, minimum=minimum, maximum=maximum
    )


def select_one(text: str, active_directories: Union[ActiveDirectory, List[ActiveDirectory]]) -> ActiveDirectory:
    """Display Active Directories in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Domain
- Host Name
- Device Count
- User Authentication Service

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __active_directories:__ any number of [ActiveDirectories](../inputs_outputs/#activedirectory)

??? success "Output"

    A single [ActiveDirectory](../inputs_outputs/#activedirectory).

??? example

    ```YAML linenums="1"
    - pam.active_directories.get_all:
        pam_address: pam.internal.net
      save: active_directories

    - pam.active_directories.select_one:
        text: Select Active Directory
      load:
        active_directories: active_directories
      save: active_directory
    ```
    """
    return select(text, active_directories, minimum=1, maximum=1)["first"]
