"""
Devices

This module contains actions for reading device information from PAM.

!!! question "Supported Versions"

    This plugin supports PAM versions 6.5.0 & newer.

!!! danger "Remember"

    You must have a PAM Appliance provisioned as a _Vault_ inside PPA to use this plugin.
"""

import uuid

from typing import List, Optional, Union

from ppa import ui
from pxm_appliance import devices, exceptions as sdk_exceptions
from playbook_tools import tables, process


from . import exceptions
from ._api import api_call, search, single_record, record_list, get_cell_value
from .records import Device, DeviceData, Secret, SecretData, Query

TABLE_HEADER = ["Name", "Host Name", "Address", "Active Directory", "Profile Count"]
TABLE_FIELDS = [
    "name",
    "hostname",
    "ip_address",
    "authentication_service_name",
    "profiles_count",
]


@api_call
def get_all(pam_address: str) -> List[Device]:
    """Get all devices from the PAM appliance.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

??? success "Output"

    A list of [Devices](../inputs_outputs/#device).

??? example
    ```YAML linenums="1"
    - pam.devices.get_all:
        pam_address: pam.internal.net
      save: all_devices
    ```
"""
    return record_list(DeviceData, devices.get_all(pam_address))


@api_call
@search
def search(pam_address: str, queries: List[Query]) -> process.SearchResult:
    """Search for devices in the PAM appliance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __queries:__ a list of [Queries](../inputs_outputs/#query) to use in the search

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [Devices](../inputs_outputs/#device) found

    - **all**: a list of [Devices](../inputs_outputs/#device) found

    - **first**: the first [Device](../inputs_outputs/#device) found

    - **last**: the last [Device](../inputs_outputs/#device) found

??? warning "Search Result Format"

    The search result format was updated in version 2.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

??? example

    Finding all __windows__ template devices in __example.domain.com__:

    ```YAML linenums="1"
    - pam.devices.search:
        pam_address: pam.internal.net
        queries:
          - attribute: device_template_name
            query: like
            value: %windows%
          - attribute: authentication_service_name
            query: equal
            value: example.domain.com
      save: search_results
    ```

!!! tip "Wildcard Queries"

    You can use the ___%___ character as a wildcard in the `value` of each query.

"""
    return process.search_result(record_list(DeviceData, devices.search(pam_address, **queries)))


@api_call
def get_by_id(pam_address: str, device_id: Union[str, int]) -> Device:
    """Get a device using its ID.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __device_id:__ the ID of the device

??? success "Output"

    A single [Device](../inputs_outputs/#device).

??? example

    ```YAML linenums="1"
    - pam.devices.get_by_id:
        pam_address: pam.internal.net
        device_id: 10
      save: device
    ```
"""
    return single_record(DeviceData, devices.get_by_id(pam_address, device_id))


@api_call
def get_by_hostname(pam_address: str, hostname: str) -> Device:
    """Get a device using its hostname.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __hostname:__ the hostname of the device

??? success "Output"

    A single [Device](../inputs_outputs/#device).

??? example

    ```YAML linenums="1"
    - pam.devices.get_by_hostname:
        pam_address: pam.internal.net
        hostname: uat-iis.internal.net
      save: uat_iis_server
    ```
"""
    try:
        return single_record(DeviceData, devices.get_by_hostname(pam_address, hostname))
    except sdk_exceptions.PxmApiError as e:
        if "column view:devices.hostname does not exist" in e.message:
            message = "Device hostnames cannot be queried in this version of PAM, this will be resolved in PAM v8.0.7."
            ui.output_error(message)
            raise exceptions.InvalidQuery(message)
        raise e


@api_call
def get_by_name(pam_address: str, name: str) -> Device:
    """Get a device using its name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __name:__ the name of the device

??? success "Output"

    A single [Device](../inputs_outputs/#device).

??? example

    ```YAML linenums="1"
    - pam.devices.get_by_name:
        pam_address: pam.internal.net
        name: uat-iis
      save: uat_iis_server
    ```
"""
    return single_record(DeviceData, devices.get_by_name(pam_address, name))


@api_call
def reveal_control_account(pam_address: str, device_id: Union[str, int]) -> Secret:
    """Reveal a device's control account secrets using its ID.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __device_id:__ the ID of the device

??? success "Output"

    A single [Secret](../inputs_outputs/#secret).

??? example

    ```YAML linenums="1"
    - pam.devices.reveal_control_account:
        pam_address: pam.internal.net
        device_id: 10
      save: device_secrets
    ```
"""
    try:
        return single_record(SecretData, devices.reveal_control_account(pam_address, device_id=device_id))
    except sdk_exceptions.PxmApiError as e:
        if e.status_code == 403:
            raise exceptions.Forbidden("Revealing the control account for this device is forbidden.")
        raise e


def output_table(text: str, devices: Union[Device, List[Device]]):
    """Display devices in a table.

The table will have the following columns:

- Name
- Host Name
- Address
- Active Directory
- Profile Count

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __devices:__ any number of [Devices](../inputs_outputs/#device)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - pam.devices.search:
        pam_address: pam.internal.net
        queries:
          - attribute: authentication_service_name
            query: equal
            value: example.domain.com
      save: domain_devices

    - pam.devices.output_table:
        text: Devices on Example Domain
      load:
        devices: domain_devices
    ```
    """
    devices = [devices] if not isinstance(devices, list) else devices
    ui.output_table(
        text, header=TABLE_HEADER, data=[[get_cell_value(device, field) for field in TABLE_FIELDS] for device in devices],
    )


def display(text: str, devices: Union[Device, List[Device]]) -> None:
    """Display devices in a table.

The table will have the following columns:

- Name
- Host Name
- Address
- Active Directory
- Profile Count

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __devices:__ any number of [Devices](../inputs_outputs/#device)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.devices.get_all:
        pam_address: pam.internal.net
      save: all_devices

    - pam.devices.display:
        text: All Devices
      load:
        devices: all_devices
    ```
    """
    tables.output_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=devices)


def input_table(
    text: str, devices: Union[Device, List[Device]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Device]:
    """Display devices in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- Host Name
- Address
- Active Directory
- Profile Count

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __devices:__ any number of [Devices](../inputs_outputs/#device)

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [Devices](../inputs_outputs/#device).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.

    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number of selections.

??? example
    ```YAML linenums="1"
    - pam.devices.search:
        pam_address: pam.internal.net
        queries:
          - attribute: authentication_service_name
            query: equal
            value: example.domain.com
      save: domain_devices

    - pam.devices.input_table:
        text: Select a Device
        minimum: 1
        maximum: 1
      load:
        devices: domain_devices
      save: selected_device
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    devices = [devices] if not isinstance(devices, list) else devices

    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[get_cell_value(device, field) for field in TABLE_FIELDS] for device in devices],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [devices[i] for i in indices]


def select(
    text: str, devices: Union[Device, List[Device]], minimum: Optional[int] = None, maximum: Optional[int] = None,
) -> tables.TableSelection:
    """Display devices in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Host Name
- Address
- Active Directory
- Profile Count

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __devices:__ any number of [Devices](../inputs_outputs/#device)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Devices](../inputs_outputs/#device)

    - **all**: a list of selected [Devices](../inputs_outputs/#device)

    - **first**: the first selected [Device](../inputs_outputs/#device)

??? example
    ```YAML linenums="1"
    - pam.devices.get_all:
        pam_address: pam.internal.net
      save: all_devices

    - pam.devices.select:
        text: Select Devices
      load:
        devices: domain_devices
      save: selection
    ```
    """
    return tables.input_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=devices, minimum=minimum, maximum=maximum)


def select_one(text: str, devices: Union[Device, List[Device]]) -> Device:
    """Display devices in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Host Name
- Address
- Active Directory
- Profile Count

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __devices:__ any number of [Devices](../inputs_outputs/#device)

??? success "Output"

    A single [Device](../inputs_outputs/#device).

??? example

    ```YAML linenums="1"
    - pam.devices.get_all:
        pam_address: pam.internal.net
      save: devices

    - pam.devices.select_one:
        text: Select Device
      load:
        devices: devices
      save: device
    ```
    """
    return select(text, devices, minimum=1, maximum=1)["first"]
