"""
Profiles

This module contains actions for reading profile information from PAM.

!!! question "Supported Versions"

    This plugin supports PAM versions 6.5.0 & newer.

!!! danger "Remember"

    You must have a PAM Appliance provisioned as a _Vault_ inside PPA to use this plugin.
"""
import uuid

from typing import List, Optional, Union

from ppa import ui
from pxm_appliance import profiles
from playbook_tools import tables, process

from ._api import api_call, search, single_record, record_list, get_cell_value
from .records import Profile, ProfileData, Query

TABLE_HEADER = ["Name", "Device Count", "User Count", "User Group Count"]
TABLE_FIELDS = ["name", "device_count", "user_count", "user_group_count"]


@api_call
def get_all(pam_address: str) -> List[Profile]:
    """Get all profiles from the PAM appliance.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

??? success "Output"

    A list of [Profiles](../inputs_outputs/#profile).

??? example
    ```YAML linenums="1"
    - pam.profiles.get_all:
        pam_address: pam.internal.net
      save: all_profiles
    ```
"""
    return record_list(ProfileData, profiles.get_all(pam_address))


@api_call
@search
def search(pam_address: str, queries: List[Query]) -> process.SearchResult:
    """Search for profiles in the PAM appliance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __queries:__ a list of [Queries](../inputs_outputs/#query) to use in the search

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [Profiles](../inputs_outputs/#profile) found

    - **all**: a list of [Profiles](../inputs_outputs/#profile) found

    - **first**: the first [Profile](../inputs_outputs/#profile) found

    - **last**: the last [Profile](../inputs_outputs/#profile) found

??? warning "Search Result Format"

    The search result format was updated in version 2.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

??? example

    Finding all profiles that contain 1 or more users:

    ```YAML linenums="1"
    - pam.profiles.search:
        pam_address: pam.internal.net
        queries:
          - attribute: user_count
            query: greater_than
            value: 0
      save: search_results
    ```

!!! tip "Wildcard Queries"

    You can use the ___%___ character as a wildcard in the `value` of each query.

"""
    return process.search_result(record_list(ProfileData, profiles.search(pam_address, **queries)))


@api_call
def get_by_id(pam_address: str, profile_id: Union[str, int]) -> Profile:
    """Get a profile using its ID.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __profile_id:__ the ID of the profile

??? success "Output"

    A single [Profile](../inputs_outputs/#profile).

??? example

    ```YAML linenums="1"
    - pam.profiles.get_by_id:
        pam_address: pam.internal.net
        profile_id: 10
      save: profile
    ```
"""
    return single_record(ProfileData, profiles.get_by_id(pam_address, profile_id))


@api_call
def get_by_name(pam_address: str, name: str) -> Profile:
    """Get a profile using its name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __name:__ the name of the device

??? success "Output"

    A single [Profile](../inputs_outputs/#profile).

??? example

    ```YAML linenums="1"
    - pam.profiles.get_by_name:
        pam_address: pam.internal.net
        name: Windows Admins
      save: profile
    ```
"""
    return single_record(ProfileData, profiles.get_by_name(pam_address, name))


def output_table(text: str, profiles: Union[Profile, List[Profile]]):
    """Display profiles in a table.

The table will have the following columns:

- Name
- Device Count
- User Count
- User Group Count

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __profiles:__ any number of [Profiles](../inputs_outputs/#profiles)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - pam.profiles.get_all:
        pam_address: pam.internal.net
      save: all_profiles

    - pam.devices.output_table:
        text: All Profiles
      load:
        profiles: all_profiles
    ```
    """
    profiles = [profiles] if not isinstance(profiles, list) else profiles
    ui.output_table(
        text, header=TABLE_HEADER, data=[[get_cell_value(profile, field) for field in TABLE_FIELDS] for profile in profiles],
    )


def display(text: str, profiles: Union[Profile, List[Profile]]) -> None:
    """Display profiles in a table.

The table will have the following columns:

- Name
- Device Count
- User Count
- User Group Count

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __profiles:__ any number of [Profiles](../inputs_outputs/#profile)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.profiles.get_all:
        pam_address: pam.internal.net
      save: profiles

    - pam.devices.display:
        text: All Profiles
      load:
        profiles: profiles
    ```
    """
    tables.output_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=profiles)


def input_table(
    text: str, profiles: Union[Profile, List[Profile]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Profile]:
    """Display profiles in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Device Count
- User Count
- User Group Count

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __profiles:__ any number of [Profiles](../inputs_outputs/#profiles)

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [Profiles](../inputs_outputs/#profile).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.

    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number of selections.

??? example
    ```YAML linenums="1"
    - pam.profiles.get_all:
        pam_address: pam.internal.net
      save: all_profiles

    - pam.devices.input_table:
        text: Select a Profile
        minimum: 1
        maximum: 1
      load:
        profiles: all_profiles
      save: selected_profile
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    profiles = [profiles] if not isinstance(profiles, list) else profiles

    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[get_cell_value(profile, field) for field in TABLE_FIELDS] for profile in profiles],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [profiles[i] for i in indices]


def select(
    text: str, profiles: Union[Profile, List[Profile]], minimum: Optional[int] = None, maximum: Optional[int] = None,
) -> tables.TableSelection:
    """Display profiles in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Device Count
- User Count
- User Group Count

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __profiles:__ any number of [Profiles](../inputs_outputs/#profile)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Profiles](../inputs_outputs/#profile)

    - **all**: a list of selected [Profiles](../inputs_outputs/#profile)

    - **first**: the first selected [Profile](../inputs_outputs/#profile)

??? example

    ```YAML linenums="1"
    - pam.profiles.get_all:
        pam_address: pam.internal.net
      save: profiles

    - pam.devices.select:
        text: Select Profiles
      load:
        profiles: profiles
      save: selection
    ```
    """
    return tables.input_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=profiles, minimum=minimum, maximum=maximum)


def select_one(text: str, profiles: Union[Profile, List[Profile]]) -> Profile:
    """Display profiles in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Device Count
- User Count
- User Group Count

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __profiles:__ any number of [Profiles](../inputs_outputs/#profile)

??? success "Output"

    A single [Profile](../inputs_outputs/#profile).

??? example

    ```YAML linenums="1"
    - pam.profiles.get_all:
        pam_address: pam.internal.net
      save: profiles

    - pam.devices.select_one:
        text: Select a Profile
      load:
        profiles: profiles
      save: profile
    ```
    """
    return select(text, profiles, minimum=1, maximum=1)["first"]
