import json

from typing import TypedDict, Optional, List, Dict, Any, Union

from datacaster.classes import CastDataClass


DateString = Optional[str]


def _cast_date_string(value):
    try:
        return value.strftime("%H:%M:%S %d/%m/%Y")
    except TypeError:
        return None


def _user_auth_type(value) -> str:
    return {
        "password": "Local",
        "local_then_radius": "Local & Radius",
        "efa": "Radius",
        "ad_ldaps": "Active Directory",
        "ad_ldaps_then_radius": "Active Directory & Radius",
    }.get(value, value)


def _notes_field(value) -> Union[str, None]:
    return value or None


class User(TypedDict):
    """
- Contains PAM user attributes

!!! info "All User Keys"

    - __auth_type__: String. One of `Local`, `Local & Radius`, `Radius`, `Active Directory`, `Active Directory & Radius`.
    - __authentication_service_id__: Integer or null. The ID of the Active Directory used for that user.
    - __authentication_service_name__: String or null. The name of the Active Directory used for that user.
    - __email_address__: String or null. The user's email address.
    - __enabled__: Boolean. Indicates whether the user is enabled.
    - __expires_at__: Date string or null. The user's expiry date.
    - __id__: Integer. The unique ID for the user.
    - __last_activity__: Date string or null. The date of the user's last PAM activity.
    - __name__: String. The user's name.
    - __notes__: String or null. The notes field for the user.
    - __profiles_count__: Integer. Number of profiles the user is in.
    - __state__: String. The state of the user account.
    - __user_groups_count__: Integer. The number of user groups the user is in.
    - __username__: String. The user's login username.
"""

    auth_type: str
    authentication_service_id: Optional[int]
    authentication_service_name: Optional[str]
    email_address: Optional[str]
    enabled: bool
    expires_at: DateString
    id: int
    last_activity: DateString
    name: str
    notes: str
    profiles_count: int
    state: str
    user_groups_count: int
    username: str


class UserGroup(TypedDict):
    """
- Contains PAM user group attributes

!!! info "All UserGroup Keys"

    - __authentication_service_id__: Integer or null. The ID of the Active Directory the group syncs with.
    - __authentication_service_name__: String or null. The name of the Active Directory the group syncs with.
    - __enabled__: Boolean. Indicates whether the user group is enabled.
    - __id__: Integer. The unique ID for the user group.
    - __is_synced_group__: Boolean. Indicates whether the user group is synced with Active Directory.
    - __name__: String. The user group's name.
    - __notes__: String or null. The notes field for the user group.
    - __profile_count__: Integer. Number of profiles the user group is in.
    - __profiles__: List of [Profile](#profile) dictionaries the user group is in.
    - __synced_at__: Date string or null. The time the user group was last synced with Active Directory.
    - __user_count__: Integer. The number of users in the user group.
"""

    authentication_service_id: Optional[int]
    authentication_service_name: Optional[str]
    enabled: bool
    id: int
    is_synced_group: bool
    name: str
    notes: str
    profile_count: int
    # Really this is a list of Profile objects, but profiles are defined further down
    # so the name isn't in scope yet. Profiles also reference user groups.
    profiles: List[Dict[str, Any]]
    synced_at: DateString
    user_count: int


class Device(TypedDict):
    """
- Contains PAM device attributes

!!! info "All Device Keys"

    - __account_source__: String. Either "Local accounts" or the name of an Active Directory the device is attached to.
    - __authentication_service_id__: Integer or null. The ID of the Active Directory the device is attached to.
    - __authentication_service_name__: String or null. The name of the Active Directory the device is attached to.
    - __control_account_id__: Integer. The ID of the Account managing the device.
    - __created_at__: Date string. The date the Active Directory was created.
    - __device_template_id__: Integer. The ID of the device template.
    - __device_template_name__: String. The name of the device template.
    - __device_vendor_name__: String. The vendor name for the device type.
    - __enabled__: Boolean. Indicates whether the device is enabled.
    - __hostname__: String. The device hostname.
    - __id__: Integer. The unique ID for the device.
    - __ip_address__: String. The device IP address or hostname.
    - __last_activity__: Date string or null. The date the device was last used in PAM.
    - __maximum_connections__: Integer or null. The maximum connection limit for the device.
    - __name__: String. The device name.
    - __notes__: String or null. The notes field for the device.
    - __profiles_count__: Integer. The number of profiles the device is in.
"""

    # Not all the fields are documented because they are rather confusing to normal
    # people. They still exist in the records for 'advanced' users (us, really).
    access_role_ids: Optional[List[int]]
    access_role_names: Optional[List[str]]
    account_source: str
    authentication_service_id: Optional[int]
    authentication_service_name: Optional[str]
    authentication_service_scheme: Optional[str]
    control_account_id: Optional[int]
    created_at: DateString
    device_template_id: int
    device_template_name: str
    device_vendor_name: str
    enabled: bool
    hostname: str
    id: int
    ip_address: str
    last_activity: DateString
    maximum_connections: Optional[int]
    name: str
    notes: str
    profiles_count: int
    traffic_light: str
    updated_at: DateString


class ActiveDirectory(TypedDict):
    """
- Contains PAM Active Directory attributes

!!! info "All UserGroup Keys"

    - __container__: String. The Organizational Unit used by PAM in Active Directory.
    - __control_account_id__:  Integer. The ID of the Account managing the Active Directory.
    - __created_at__: Date string. The date the Active Directory was created.
    - __device_count__: Integer. The number of devices attached to the Active Directory.
    - __domain__: String. The Active Directory fqdn.
    - __enabled__: Integer. Indicates whether the Active Directory is enabled.
    - __groups__: List of strings. The Security Groups accounts are audited from. If empty, all accounts are audited.
    - __hostname__: String. The IP address or hostname of the domain controller(s) used by PAM.
    - __id__: Integer. The unique ID for the Active Directory.
    - __name__: String. The Active Directory name.
    - __notes__: String or null. The notes field for the Active Directory.
    - __user_authentication_service__: Boolean. Indicates whether this domain is used to authentication users into PAM.
"""

    container: str
    control_account_id: Optional[int]
    created_at: DateString
    device_count: int
    domain: str
    enabled: bool
    groups: str  # this is a JSON string containing a list of user groups for some reason!
    hostname: str
    id: int
    name: str
    notes: str
    updated_at: DateString
    user_authentication_service: bool


class StaticVault(TypedDict):
    """
- Contains PAM static vault attributes

!!! info "All UserGroup Keys"

    - __created_at__: Date string. The date the static vault was created.
    - __device_count__: Integer. The number of devices attached to the static vault.
    - __id__: Integer. The unique ID for the static vault.
    - __name__: String. The static vault name.
"""

    device_count: int
    created_at: DateString
    id: int
    name: str


class Profile(TypedDict):
    """
- Contains PAM profile attributes

!!! info "All Profile Keys"

    - __device_count__: Integer. The number of devices in the profile.
    - __groups__: List of [UserGroup](#usergroup) dictionaries in the profile.
    - __id__: Integer. The unique ID for the profile.
    - __name__: String. The profile name.
    - __schedule_count__: Integer. The number of schedules assigned to tasks in the profile.
    - __task_count__: Integer. The number of tasks in the profile.
    - __tool_count__: Integer. The number of tools in the profile.
    - __user_count__: Integer. The number of users directly in the profile.
    - __user_group_count__: Integer. The number of user groups in the profile.
"""

    device_count: int
    groups: List[UserGroup]  # This is a list of UserGroup objects.
    id: int
    name: str
    schedule_count: int
    state: str
    task_count: int
    tool_count: int
    user_count: int
    user_group_count: int


class Account(TypedDict):
    """
- Contains PAM account attributes

!!! info "All Account Keys"

    - __created_at__: Date string. The date the account was either created or initially audited.
    - __credentials_last_changed_at__: Date string or null. The date the credentials were last changed by PAM.
    - __has_password__: Boolean. Indicates whether the account has a password known by PAM.
    - __has_sshkey__: Boolean. Indicates whether the account has an SSH key known by PAM.
    - __id__: Integer. The unique ID for the account.
    - __name__: String. The name of the account.
    - __parent_id__: Integer. The ID of the Active Directory, device, or static vault the account is attached to.
    - __parent_type__: String. One of "devices", "active-directories", or "static-vaults".
    - __state__: String. The account state. One of "Unapproved", "Approved", "Known", "Managed", "Fully Managed".
"""

    created_at: DateString
    credentials_last_changed_at: Optional[DateString]
    has_password: bool
    has_sshkey: bool
    id: int
    invalid_at: DateString
    locked: bool
    name: str
    parent_id: int
    parent_type: str
    state: str
    updated_at: DateString


class SSHKey(TypedDict):
    """
- Contains an SSH key & passphrase linked to an account.

- Used in [secret](#secret) dictionaries if an account has an SSH key.

!!! info "All Secret Keys"

    - __private__: The SSH private key.
    - __passphrase__: The SSH private key passphrase.
"""

    passphrase: Optional[str]
    private: Optional[str]


class Secret(TypedDict):
    """
- Contains an account's secrets.

- Used as an output from the __reveal__ actions in this plugin.

!!! info "All Secret Keys"

    - __account_id__: The ID of the [account](#account).
    - __account_usename__: The account username.
    - __password__: The account password (if applicable).
    - __ssh_key__: The account [SSH key](#sshkey) (if applicable).
"""

    account_id: int
    account_username: str
    password: Optional[str]
    ssh_key: Optional[SSHKey]


class Query(TypedDict):
    """
- Contains a search query & value.

- Used as an input to the __search__ actions in this plugin.

!!! example "Searching with _like_ & a wildcard"

    ```YAML linenums="1"
    - pam.devices.search:
        pam_address: internal.pam.net
        queries:
          - attribute: hostname
            query: like
            value: "%-uat"
    ```

!!! example "Searching with _not_ & __null__"

    ```YAML linenums="1"
    - pam.devices.search:
        pam_address: internal.pam.net
        queries:
          - attribute: authentication_service_id
            query: not
            value: null
    ```

!!! warning "Required Keys"

    - __attribute__: The name of the attribute to query against.
    - __query__: The query type (see valid query types below).
    - __value__: The value to use in the query. Use ___%___ as a wildcard.

!!! tip "Valid Query Types"

    - `equal` (_string_ or _integer_)
    - `not_equal` (_string_ or _integer_)
    - `like` (_string_)
    - `greater_than` (_integer_)
    - `less_than` (_integer_)
    - `is` (_null_)
    - `not` (_null_)

"""

    attribute: str
    query: str
    value: Union[str, int, bool, None]


# These are at the bottom so they can reference all classes above.


class UserData(CastDataClass):
    __annotations__ = User.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "types": {DateString: _cast_date_string},
            "fields": {"notes": _notes_field, "auth_type": _user_auth_type},
        },
        # This is always cast because the incoming type is a string & won't fail type-checking.
        "always_cast": ["auth_type"],
    }


class DeviceData(CastDataClass):
    __annotations__ = Device.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {"types": {DateString: _cast_date_string}, "fields": {"notes": _notes_field}},
    }


class ActiveDirectoryData(CastDataClass):
    __annotations__ = ActiveDirectory.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "types": {DateString: _cast_date_string},
            "fields": {
                "notes": _notes_field,
                "groups": lambda groups: [d["value"] for d in json.loads(groups)] if groups else [],
            },
        },
    }


class StaticVaultData(CastDataClass):
    __annotations__ = StaticVault.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {"types": {DateString: _cast_date_string}},
    }


class ProfileData(CastDataClass):
    __annotations__ = Profile.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "types": {DateString: _cast_date_string},
            "fields": {"groups": lambda groups: [UserGroup(**group).__dict__ for group in groups] if groups else []},
        },
    }


class UserGroupData(CastDataClass):
    __annotations__ = UserGroup.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "types": {DateString: _cast_date_string},
            "fields": {
                "notes": _notes_field,
                "profiles": lambda profiles: [ProfileData(**p).__dict__ for p in profiles] if profiles else [],
            },
        },
    }


class AccountData(CastDataClass):
    __annotations__ = Account.__dict__["__annotations__"]
    __class_config__ = {"cast_functions": {"fields": {"state": lambda state: state.replace("_", " ").title()}}}


class SecretData(CastDataClass):
    __annotations__ = Secret.__dict__["__annotations__"]
    __class_config__ = {"cast_functions": {"fields": {"ssh_key": lambda ssh_key: SSHKey(**ssh_key.__dict__)}}}
