"""
Static Vaults

This module contains actions for reading static vault information from PAM.

!!! question "Supported Versions"

    This plugin supports PAM versions 6.5.0 & newer.

!!! danger "Remember"

    You must have a PAM Appliance provisioned as a _Vault_ inside PPA to use this plugin.
"""

import uuid

from typing import List, Optional, Union

from ppa import ui
from pxm_appliance import static_vaults
from playbook_tools import tables, process

from ._api import api_call, search, single_record, record_list, get_cell_value
from .records import StaticVault, StaticVaultData, Secret, SecretData, Query


TABLE_HEADER = ["Name", "Device Count"]
TABLE_FIELDS = ["name", "device_count"]


@api_call
def get_all(pam_address: str) -> List[StaticVault]:
    """Get all static vaults from the PAM appliance.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

??? success "Output"

    A list of [StaticVaults](../inputs_outputs/#staticvault).

??? example
    ```YAML linenums="1"
    - pam.static_vaults.get_all:
        pam_address: pam.internal.net
      save: all_static_vaults
    ```
"""
    return record_list(StaticVaultData, static_vaults.get_all(pam_address))


@api_call
@search
def search(pam_address: str, queries: List[Query]) -> process.SearchResult:
    """Search for static vaults in the PAM appliance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __queries:__ a list of [Queries](../inputs_outputs/#query) to use in the search

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [StaticVaults](../inputs_outputs/#staticvault) found

    - **all**: a list of [StaticVaults](../inputs_outputs/#staticvault) found

    - **first**: the first [StaticVault](../inputs_outputs/#staticvault) found

    - **last**: the last [StaticVault](../inputs_outputs/#staticvault) found

??? warning "Search Result Format"

    The search result format was updated in version 2.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

??? example

    Find all static vaults with 1 or more devices:

    ```YAML linenums="1"
    - pam.static_vaults.search:
        pam_address: pam.internal.net
        queries:
          - attribute: device_count
            query: greater_than
            value: 0
      save: search_results
    ```

!!! tip "Wildcard Queries"

    You can use the ___%___ character as a wildcard in the `value` of each query.

"""
    return process.search_result(record_list(StaticVaultData, static_vaults.search(pam_address, **queries)))


@api_call
def get_by_id(pam_address: str, vault_id: Union[str, int]) -> StaticVault:
    """Get a static vault using its ID.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __vault_id:__ the ID of the Static Vault

??? success "Output"

    A single [StaticVault](../inputs_outputs/#staticvault).

??? example

    ```YAML linenums="1"
    - pam.static_vaults.get_by_id:
        pam_address: pam.internal.net
        vault_id: 2
      save: static_vault
    ```
"""
    return single_record(StaticVaultData, static_vaults.get_by_id(pam_address, vault_id))


@api_call
def get_by_name(pam_address: str, name: str) -> StaticVault:
    """Get a static vault using its name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __name:__ the name of the Static Vault

??? success "Output"

    A single [StaticVault](../inputs_outputs/#staticvault).

??? example

    ```YAML linenums="1"
    - pam.static_vaults.get_by_name:
        pam_address: pam.internal.net
        name: Cloud Secrets
      save: static_vault
    ```
"""
    return single_record(StaticVaultData, static_vaults.get_by_name(pam_address, name))


@api_call
def reveal_account_by_name(pam_address: str, vault_id: Union[str, int], account_name: str) -> Secret:
    """Reveal a static vault account's secrets using the account name.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __vault_id:__ the ID of the Static Vault

    __account_name:__ the name of the Static Vault account

??? success "Output"

    A single [Secret](../inputs_outputs/#secret).

??? example

    ```YAML linenums="1"
    - pam.static_vaults.reveal_account_by_name:
        pam_address: pam.internal.net
        vault_id: 3
        account_name: api_key
      save: secret
    ```
"""
    return single_record(
        SecretData, static_vaults.reveal_account_by_name(pam_address, vault_id=vault_id, account_name=account_name)
    )


def output_table(text: str, static_vaults: Union[StaticVault, List[StaticVault]]):
    """Display static vaults in a table.

The table will have the following columns:

- Name
- Device Count

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __static_vaults:__ any number of [StaticVaults](../inputs_outputs/#staticvault)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.static_vaults.get_all:
        pam_address: pam.internal.net
      save: all_static_vaults

    - pam.static_vaults.output_table:
        text: All Static Vaults
      load:
        static_vaults: all_static_vaults
    ```
    """
    static_vaults = [static_vaults] if not isinstance(static_vaults, list) else static_vaults
    ui.output_table(
        text, header=TABLE_HEADER, data=[[get_cell_value(vault, field) for field in TABLE_FIELDS] for vault in static_vaults],
    )


def display(text: str, static_vaults: Union[StaticVault, List[StaticVault]]) -> None:
    """Display static vaults in a table.

The table will have the following columns:

- Name
- Device Count

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __static_vaults:__ any number of [StaticVaults](../inputs_outputs/#staticvault)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.static_vaults.get_all:
        pam_address: pam.internal.net
      save: static_vaults

    - pam.static_vaults.display:
        text: Static Vaults
      load:
        static_vaults: static_vaults
    ```
    """
    tables.output_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=static_vaults)


def input_table(
    text: str, static_vaults: Union[StaticVault, List[StaticVault]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[StaticVault]:
    """Display static vaults in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Device Count

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __static_vaults:__ any number of [StaticVaults](../inputs_outputs/#staticvault)

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [StaticVaults](../inputs_outputs/#staticvault).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.

    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number of selections.

??? example

    ```YAML linenums="1"
    - pam.static_vaults.get_all:
        pam_address: pam.internal.net
      save: static_vaults

    - pam.static_vaults.input_table:
        text: Select a Static Vault
        minimum: 1
        maximum: 1
      load:
        static_vaults: static_vaults
      save: selected_static_vaults
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    static_vaults = [static_vaults] if not isinstance(static_vaults, list) else static_vaults

    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[get_cell_value(vault, field) for field in TABLE_FIELDS] for vault in static_vaults],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [static_vaults[i] for i in indices]


def select(
    text: str, static_vaults: Union[StaticVault, List[StaticVault]], minimum: Optional[int] = None, maximum: Optional[int] = None,
) -> tables.TableSelection:
    """Display static vaults in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Device Count

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __static_vaults:__ any number of [StaticVaults](../inputs_outputs/#staticvault)

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [StaticVaults](../inputs_outputs/#staticvault)

    - **all**: a list of selected [StaticVaults](../inputs_outputs/#staticvault)

    - **first**: the first selected [StaticVault](../inputs_outputs/#staticvault)

??? example

    ```YAML linenums="1"
    - pam.static_vaults.get_all:
        pam_address: pam.internal.net
      save: static_vaults

    - pam.static_vaults.select:
        text: Select Static Vaults
      load:
        static_vaults: static_vaults
      save: selection
    ```
    """
    return tables.input_table(
        text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=static_vaults, minimum=minimum, maximum=maximum
    )


def select_one(text: str, static_vaults: Union[StaticVault, List[StaticVault]],) -> tables.TableSelection:
    """Display static vaults in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Device Count

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __static_vaults:__ any number of [StaticVaults](../inputs_outputs/#staticvault)

??? success "Output"

    A single [StaticVault](../inputs_outputs/#staticvault).

??? example

    ```YAML linenums="1"
    - pam.static_vaults.get_all:
        pam_address: pam.internal.net
      save: static_vaults

    - pam.static_vaults.select_one:
        text: Select Static Vault
      load:
        static_vaults: static_vaults
      save: static_vault
    ```
    """
    return select(text, static_vaults, minimum=1, maximum=1)["first"]
