"""
User Groups

This module contains actions for reading user group information from PAM.

!!! question "Supported Versions"

    This plugin supports PAM versions 6.5.0 & newer.

!!! danger "Remember"

    You must have a PAM Appliance provisioned as a _Vault_ inside PPA to use this plugin.
"""

import uuid
import logging

from typing import List, Union, Optional, Any

from ppa import ui
from pxm_appliance import user_groups
from playbook_tools import tables, process

from ._api import api_call, search, single_record, record_list, get_cell_value
from .records import Profile, ProfileData, UserGroup, UserGroupData, User, UserData, Query
from .exceptions import NoSuchObject


TABLE_HEADER = ["Name", "AD Synchronised", "Profile Count", "User Count", "Enabled"]
TABLE_FIELDS = ["name", "is_synced_group", "profile_count", "user_count", "enabled"]

logger = logging.getLogger(__name__)


@api_call
def get_all(pam_address: str) -> List[UserGroup]:
    """Get all user groups from the PAM appliance.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

??? success "Output"

    A list of [UserGroups](../inputs_outputs/#usergroup).

??? example
    ```YAML linenums="1"
    - pam.user_groups.get_all:
        pam_address: pam.internal.net
      save: all_user_groups
    ```
"""
    return record_list(UserGroupData, user_groups.get_all(pam_address))


@api_call
@search
def search(pam_address: str, queries: List[Query]) -> process.SearchResult:
    """Search for user groups in the PAM appliance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __queries:__ a list of [Queries](../inputs_outputs/#query) to use in the search

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [UserGroups](../inputs_outputs/#usergroup) found

    - **all**: a list of [UserGroups](../inputs_outputs/#usergroup) found

    - **first**: the first [UserGroup](../inputs_outputs/#usergroup) found

    - **last**: the last [UserGroup](../inputs_outputs/#usergroup) found

??? warning "Search Result Format"

    The search result format was updated in version 2.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

??? example

    Find all user groups with 1 or more users:

    ```YAML linenums="1"
    - pam.user_groups.search:
        pam_address: pam.internal.net
        queries:
          - attribute: user_count
            query: greater_than
            value: 0
      save: search_results
    ```

!!! tip "Wildcard Queries"

    You can use the ___%___ character as a wildcard in the `value` of each query.

"""
    return process.search_result(record_list(UserGroupData, user_groups.search(pam_address, **queries)))


@api_call
def get_by_id(pam_address: str, user_group_id: Union[str, int]) -> UserGroup:
    """Get a user group using its ID.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_group_id:__ the ID of the user group

??? success "Output"

    A single [UserGroup](../inputs_outputs/#usergroup).

??? example

    ```YAML linenums="1"
    - pam.user_groups.get_by_id:
        pam_address: pam.internal.net
        user_group_id: 2
      save: user_group
    ```
"""
    return single_record(UserGroupData, user_groups.get_by_id(pam_address, user_group_id))


@api_call
def get_by_name(pam_address: str, name: str) -> UserGroup:
    """Get a user group using its name.

!!! question "Minimum Plugin Version: 3.2.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __name:__ the name of the user group

??? success "Output"

    A single [UserGroup](../inputs_outputs/#usergroup).

??? example

    ```YAML linenums="1"
    - pam.user_groups.get_by_name:
        pam_address: pam.internal.net
        user_group_id: Admins
      save: user_group
    ```
"""
    if result := search(pam_address, queries=[{"attribute": "name", "query": "equal", "value": name}])["first"]:
        return result
    raise NoSuchObject(f"No user group found with name {name}.")


@api_call
def get_profiles(pam_address: str, user_group_id: Union[str, int]) -> List[Profile]:
    """Get all profiles a particular user group is assigned to.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_group_id:__ the ID of the user group

??? success "Output"

    A list of [Profiles](../inputs_outputs/#profile).

??? example

    ```YAML linenums="1"
    - pam.user_groups.get_profiles:
        pam_address: pam.internal.net
        user_group_id: 2
      save: profiles
    ```
"""
    return record_list(ProfileData, user_groups.get_profiles(pam_address, user_group_id=user_group_id))


def in_profile(pam_address: str, user_group_id: Union[str, int], profile_id: Union[int, str]) -> bool:
    """Determine if the supplied user group is in a particular profile.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_group_id:__ the ID of the user group

    __profile_id:__ the ID of the profile to check

??? success "Output"

    A boolean is outputted by this action:

    - `true` if the supplied user group is in the profile

    - `false` if the supplied user group is not in the group

??? example

    ```YAML linenums="1"
    - pam.user_groups.in_profile:
        pam_address: pam.internal.net
        user_group_id: 2
        profile_id: 10
      save: in_profile
    ```
"""
    return profile_id in list(map(lambda p: p["id"], get_profiles(pam_address, user_group_id)))


@api_call
def get_users(pam_address: str, user_group_id: Union[str, int]) -> List[User]:
    """Get all users assigned to a particular user group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_group_id:__ the ID of the user group

??? success "Output"

    A list of [Users](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - pam.user_groups.get_users:
        pam_address: pam.internal.net
        user_group_id: 2
      save: users
    ```
"""
    return record_list(UserData, user_groups.get_users(pam_address, user_group_id=user_group_id))


def generate_report(pam_address: str, appliance_name: Optional[str] = None) -> List[List[Any]]:
    """Generate a user group report showing each user membership for each group.

The table can be displayed in the task interface with the [display_report](#display_report) action.

You can combine reports from multiple appliances before displaying them (see examples below).

!!! question "Minimum Plugin Version: 3.3.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __appliance_name:__ a friendly name for the PAM appliance in the report (defaults to `pam_address`)

??? success "Output"

    A report containing each user group membership on the supplied PAM Appliance.

??? example "Example 1 - Single Appliance"

    ```YAML linenums="1"
    - pam.user_groups.generate_report:
        pam_address: 192.168.123.4
        appliance_name: Internal PAM Appliance
      save: report

    - pam.user_groups.display_report:
        text: Internal PAM User Group Report
      load:
        report: report
    ```

??? example "Example 2 - Multiple Appliances"

    ```YAML linenums="1" hl_lines="5 11 14 15 16 17 18 24"
    # Generate a report from the first appliance
    - pam.user_groups.generate_report:
        pam_address: 192.168.123.4
        appliance_name: Internal PAM Appliance
      save: internal_report

    # Generate a report from the second appliance
    - pam.user_groups.generate_report:
        pam_address: 192.168.123.5
        appliance_name: Third-Party PAM Appliance
      save: third_party_report

    # Combine both reports into a single report
    - ppa_tools.lists.combine:
      load:
        first: internal_report
        second: third_party_report
      save: report

    # Display the combined report
    - pam.user_groups.display_report:
        text: PAM User Group Reports
      load:
        report: report
    ```
"""
    appliance_name = appliance_name or pam_address
    rows = []
    for group in get_all(pam_address):
        for user in get_users(pam_address, group["id"]):
            rows.append([appliance_name, group["name"], user["username"], user["name"]])
    return rows


def display_report(text: str, report: List[List[Any]]) -> None:
    """Display a table generated by the [generate_report](#generate_report) action.

You can combine reports from multiple appliances before displaying them (see [generate_report](#generate_report)).

The table will have the following columns:

- Appliance
- User Group
- User Name
- User Display Name

!!! question "Minimum Plugin Version: 3.3.0"

??? info "Input"

    __text:__ the title of the table

    __report:__ the output of [generate_report](#generate_report)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.user_groups.generate_report:
        pam_address: 192.168.123.4
        appliance_name: Internal PAM Appliance
      save: report

    - pam.user_groups.display_report:
        text: Internal PAM User Group Report
      load:
        report: report
    ```
"""
    ui.output_table(text, header=["Appliance", "User Group", "User Name", "User Display Name"], data=report)


def output_table(text: str, groups: Union[UserGroup, List[UserGroup]]):
    """Display user groups in a table.

The table will have the following columns:

- Name
- AD Synchronised
- Profile Count
- User Count
- Enabled

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ a single or list of [UserGroups](../inputs_outputs/#usergroup) to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.user_groups.get_all:
        pam_address: pam.internal.net
      save: all_user_groups

    - pam.user_groups.output_table:
        text: All User Groups
      load:
        groups: all_user_groups
    ```
    """
    ui.output_table(
        text, header=TABLE_HEADER, data=[[get_cell_value(group, field) for field in TABLE_FIELDS] for group in groups]
    )


def display(text: str, user_groups: Union[UserGroup, List[UserGroup]]) -> None:
    """Display user groups in a table.

The table will have the following columns:

- Name
- AD Synchronised
- Profile Count
- User Count
- Enabled

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [UserGroups](../inputs_outputs/#usergroup)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.user_groups.get_all:
        pam_address: pam.internal.net
      save: user_groups

    - pam.user_groups.display:
        text: User Groups
      load:
        user_groups: user_groups
    ```
    """
    tables.output_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=user_groups)


def input_table(
    text: str, groups: Union[UserGroup, List[UserGroup]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[UserGroup]:
    """Display user groups in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- AD Synchronised
- Profile Count
- User Count
- Enabled

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __groups:__ a single or list of [UserGroups](../inputs_outputs/#usergroup) to display in the table

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [UserGroups](../inputs_outputs/#usergroup).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.

    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number of selections.

??? example

    ```YAML linenums="1"
    - pam.user_groups.get_all:
        pam_address: pam.internal.net
      save: all_user_groups

    - pam.user_groups.input_table:
        text: Select a User Group
        minimum: 1
        maximum: 1
      load:
        groups: all_user_groups
      save: selected_user_groups
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    groups = [groups] if not isinstance(groups, list) else groups

    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[get_cell_value(group, field) for field in TABLE_FIELDS] for group in groups],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [groups[i] for i in indices]


def select(
    text: str, user_groups: Union[UserGroup, List[UserGroup]], minimum: Optional[int] = None, maximum: Optional[int] = None,
) -> tables.TableSelection:
    """Display user groups in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- AD Synchronised
- Profile Count
- User Count
- Enabled

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __user_groups:__ any number of [UserGroups](../inputs_outputs/#usergroup)

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [UserGroups](../inputs_outputs/#usergroup)

    - **all**: a list of selected [UserGroups](../inputs_outputs/#usergroup)

    - **first**: the first selected [UserGroup](../inputs_outputs/#usergroup)

??? example

    ```YAML linenums="1"
    - pam.user_groups.get_all:
        pam_address: pam.internal.net
      save: user_groups

    - pam.user_groups.select:
        text: Select User Groups
      load:
        user_groups: user_groups
      save: selection
    ```
    """
    return tables.input_table(
        text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=user_groups, minimum=minimum, maximum=maximum
    )


def select_one(text: str, user_groups: Union[UserGroup, List[UserGroup]],) -> tables.TableSelection:
    """Display user groups in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- AD Synchronised
- Profile Count
- User Count
- Enabled

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __user_groups:__ any number of [UserGroups](../inputs_outputs/#usergroup)

??? success "Output"

    A single [UserGroup](../inputs_outputs/#usergroup).

??? example

    ```YAML linenums="1"
    - pam.user_groups.get_all:
        pam_address: pam.internal.net
      save: user_groups

    - pam.user_groups.select_one:
        text: Select User Group
      load:
        user_groups: user_groups
      save: user_group
    ```
    """
    return select(text, user_groups, minimum=1, maximum=1)["first"]
