"""
Users

This module contains actions for reading user information from PAM.

!!! question "Supported Versions"

    This plugin supports PAM versions 6.5.0 & newer.

!!! danger "Remember"

    You must have a PAM Appliance provisioned as a _Vault_ inside PPA to use this plugin.
"""
import uuid

from typing import List, Union, Optional

from ppa import ui
from pxm_appliance import users
from playbook_tools import tables, process

from ._api import api_call, search, single_record, record_list, get_cell_value
from .records import User, UserData, Device, DeviceData, UserGroup, UserGroupData, Profile, ProfileData, Query


TABLE_HEADER = ["Name", "User Name", "Email Address", "Authentication Type", "Enabled", "Expires At"]
TABLE_FIELDS = ["name", "username", "email_address", "auth_type", "enabled", "expires_at"]


@api_call
def get_all(pam_address: str) -> List[User]:
    """Get all users from the PAM appliance.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

??? success "Output"

    A list of [Users](../inputs_outputs/#user).

??? example
    ```YAML linenums="1"
    - pam.users.get_all:
        pam_address: pam.internal.net
      save: all_users
    ```
"""
    return record_list(UserData, users.get_all(pam_address))


@api_call
@search
def search(pam_address: str, queries: List[Query]) -> process.SearchResult:
    """Search for users in the PAM appliance.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __queries:__ a list of [Queries](../inputs_outputs/#query) to use in the search

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [Users](../inputs_outputs/#user) found

    - **all**: a list of [Users](../inputs_outputs/#user) found

    - **first**: the first [User](../inputs_outputs/#user) found

    - **last**: the last [User](../inputs_outputs/#user) found

??? warning "Search Result Format"

    The search result format was updated in version 2.0.0 of this plugin.

    It is not compatible with playbooks written again previous versions.

    See [here](/searches/#using-search-results) for more information.

??? example

    Find all disabled users:

    ```YAML linenums="1"
    - pam.users.search:
        pam_address: pam.internal.net
        queries:
          - attribute: enabled
            query: is
            value: false
      save: search_results
    ```

!!! tip "Wildcard Queries"

    You can use the ___%___ character as a wildcard in the `value` of each query.

"""
    return process.search_result(record_list(UserData, users.search(pam_address, **queries)))


@api_call
def get_by_id(pam_address: str, user_id: Union[str, int]) -> User:
    """Get a user using its ID.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_id:__ the ID of the user

??? success "Output"

    A single [User](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - pam.users.get_by_id:
        pam_address: pam.internal.net
        user_id: 2
      save: user
    ```
"""
    return single_record(UserData, users.get_by_id(pam_address, user_id))


@api_call
def get_by_username(pam_address: str, username: str) -> User:
    """Get a user using its username.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __username:__ the user's username

??? success "Output"

    A single [User](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - pam.users.get_by_username:
        pam_address: pam.internal.net
        username: example.user
      save: user
    ```
"""
    return single_record(UserData, users.get_by_username(pam_address, username))


@api_call
def get_devices(pam_address: str, user_id: Union[str, int]) -> List[Device]:
    """Get all devices the supplied user has access to.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_id:__ the ID of the user

??? success "Output"

    A list of [Devices](../inputs_outputs/#device).

??? example

    ```YAML linenums="1"
    - pam.users.get_devices:
        pam_address: pam.internal.net
        user_id: 2
      save: devices
    ```
"""
    return record_list(DeviceData, users.get_all_devices(pam_address, user_id=user_id))


@api_call
def get_user_groups(pam_address: str, user_id: Union[str, int]) -> List[UserGroup]:
    """Get all user groups the supplied user is assigned to.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_id:__ the ID of the user

??? success "Output"

    A list of [UserGroups](../inputs_outputs/#usergroup).

??? example

    ```YAML linenums="1"
    - pam.users.get_user_groups:
        pam_address: pam.internal.net
        user_id: 2
      save: user_groups
    ```
"""
    return record_list(UserGroupData, users.get_user_groups(pam_address, user_id=user_id))


@api_call
def get_direct_profiles(pam_address: str, user_id: Union[str, int]) -> List[Profile]:
    """Get all profiles the supplied user is _directly_ assigned to.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_id:__ the ID of the user

??? success "Output"

    A list of [Profiles](../inputs_outputs/#profile).

??? example

    ```YAML linenums="1"
    - pam.users.get_direct_profiles:
        pam_address: pam.internal.net
        user_id: 2
      save: direct_profiles
    ```
"""
    return record_list(ProfileData, users.get_direct_profiles(pam_address, user_id=user_id))


@api_call
def get_user_group_profiles(pam_address: str, user_id: Union[str, int]) -> List[Profile]:
    """Get all profiles the supplied user is assigned to via a _user group_.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_id:__ the ID of the user

??? success "Output"

    A list of [Profiles](../inputs_outputs/#profile).

??? example

    ```YAML linenums="1"
    - pam.users.get_user_group_profiles:
        pam_address: pam.internal.net
        user_id: 2
      save: user_group_profiles
    ```
"""
    return record_list(ProfileData, users.get_user_group_profiles(pam_address, user_id=user_id))


@api_call
def get_all_profiles(pam_address: str, user_id: Union[str, int]) -> List[Profile]:
    """Get all profiles the supplied user is assigned to.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_id:__ the ID of the user

??? success "Output"

    A list of [Profiles](../inputs_outputs/#profile).

??? example

    ```YAML linenums="1"
    - pam.users.get_all_profiles:
        pam_address: pam.internal.net
        user_id: 2
      save: profiles
    ```
"""
    return record_list(ProfileData, users.get_all_profiles(pam_address, user_id=user_id))


def in_profile(pam_address: str, user_id: Union[str, int], profile_id: Union[int, str]) -> bool:
    """Determine if the supplied user is in a particular profile.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    __pam_address:__ the IP or DNS address of the PAM Appliance

    __user_id:__ the ID of the user group

    __profile_id:__ the ID of the profile to check

??? success "Output"

    A boolean is outputted by this action:

    - `true` if the supplied user group is in the profile

    - `false` if the supplied user group is not in the group

??? example

    ```YAML linenums="1"
    - pam.users.in_profile:
        pam_address: pam.internal.net
        user_id: 2
        profile_id: 10
      save: in_profile
    ```
"""
    return profile_id in list(map(lambda p: p["id"], get_all_profiles(pam_address, user_id)))


def output_table(text: str, users: Union[User, List[User]]):
    """Display users in a table.

The table will have the following columns:

- Name
- Username
- Email Address
- Authentication Type
- Enabled
- Expires At

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.users.get_all:
        pam_address: pam.internal.net
      save: all_users

    - pam.users.output_table:
        text: All Users
      load:
        groups: all_users
    ```
    """
    users = [users] if not isinstance(users, list) else users
    ui.output_table(
        text, header=TABLE_HEADER, data=[[get_cell_value(user, field) for field in TABLE_FIELDS] for user in users],
    )


def display(text: str, users: Union[User, List[User]]) -> None:
    """Display users in a table.

The table will have the following columns:

- Name
- Username
- Email Address
- Authentication Type
- Enabled
- Expires At

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - pam.users.get_all:
        pam_address: pam.internal.net
      save: users

    - pam.users.display:
        text: Users
      load:
        users: users
    ```
    """
    tables.output_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=users)


def input_table(
    text: str, users: Union[User, List[User]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[User]:
    """Display users in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Name
- Username
- Email Address
- Authentication Type
- Enabled
- Expires At

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __users:__ a single or list of [Users](../inputs_outputs/#user) to display in the table

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A list of [Users](../inputs_outputs/#users).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.

    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number of selections.

??? example

    ```YAML linenums="1"
    - pam.users.get_all:
        pam_address: pam.internal.net
      save: all_users

    - pam.users.input_table:
        text: Select a User
        minimum: 1
        maximum: 1
      load:
        users: all_users
      save: selected_users
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    users = [users] if not isinstance(users, list) else users

    while True:
        indices = ui.input_table(
            text,
            header=TABLE_HEADER,
            data=[[get_cell_value(user, field) for field in TABLE_FIELDS] for user in users],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [users[i] for i in indices]


def select(
    text: str, users: Union[User, List[User]], minimum: Optional[int] = None, maximum: Optional[int] = None,
) -> tables.TableSelection:
    """Display users in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Username
- Email Address
- Authentication Type
- Enabled
- Expires At

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

    - __minimum:__ The minimum number of acceptable selections

    - __maximum:__ The maximum number of acceptable selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Users](../inputs_outputs/#user)

    - **all**: a list of selected [Users](../inputs_outputs/#user)

    - **first**: the first selected [User](../inputs_outputs/#user)

??? example

    ```YAML linenums="1"
    - pam.users.get_all:
        pam_address: pam.internal.net
      save: users

    - pam.users.select:
        text: Select Users
      load:
        users: users
      save: selection
    ```
    """
    return tables.input_table(text, header=TABLE_HEADER, extract=TABLE_FIELDS, records=users, minimum=minimum, maximum=maximum)


def select_one(text: str, users: Union[User, List[User]],) -> tables.TableSelection:
    """Display users in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Username
- Email Address
- Authentication Type
- Enabled
- Expires At

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __text:__ The title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

??? success "Output"

    A single [User](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - pam.users.get_all:
        pam_address: pam.internal.net
      save: users

    - pam.users.select_one:
        text: Select User
      load:
        users: users
      save: user
    ```
    """
    return select(text, users, minimum=1, maximum=1)["first"]
